import os
from textwrap import dedent

import numpy as np
import h5py

from fluiddyn.util import mpi

from .base import SpecificOutput
from .movies import MoviesBase1D


class MoviesSpectra(MoviesBase1D):
    def __init__(self, output, spectra):
        self.spectra = spectra
        super().__init__(output)

    def init_animation(self, *args, **kwargs):
        if "xmax" not in kwargs:
            kwargs["xmax"] = self.oper.k_spectra3d[-1]
        if "ymax" not in kwargs:
            kwargs["ymax"] = 1.0

        with h5py.File(self.spectra.path_file3d) as file:
            self.times = file["times"][...]

        super().init_animation(*args, **kwargs)

    def get_field_to_plot(self, time, key=None):
        if key is None:
            key = self.key_field
        idx, t_file = self.get_closest_time_file(time)
        with h5py.File(self.spectra.path_file3d) as file:
            y = file["spectra_" + key][idx]
        y[abs(y) < 10e-16] = 0
        return y, t_file

    def get_closest_time_file(self, time):
        """Find the index and value of the closest actual time of the field."""
        idx = np.abs(self.times - time).argmin()
        return idx, self.times[idx]

    def _init_labels(self, xlabel="x"):
        """Initialize the labels."""
        self.ax.set_xlabel(xlabel, fontdict=self.font)
        self.ax.set_ylabel(self.key_field, fontdict=self.font)
        self.ax.set_yscale("log")

    def _get_axis_data(self):
        """Get axis data.

        Returns
        -------

        x : array
          x-axis data.

        """
        with h5py.File(self.spectra.path_file3d) as file:
            x = file["k_spectra3d"][...]

        return x

    def _set_key_field(self, key_field):
        """
        Defines key_field default.
        """
        if key_field is None:
            key_field = "E"
        self.key_field = key_field


class Spectra(SpecificOutput):
    """Used for the saving of spectra."""

    _tag = "spectra"

    @staticmethod
    def _complete_params_with_default(params):
        tag = "spectra"

        params.output.periods_save._set_attrib(tag, 0)
        p_spectra = params.output._set_child(
            tag, attribs={"HAS_TO_PLOT_SAVED": False, "kzkh_periodicity": 0}
        )
        p_spectra._set_doc(
            dedent(
                """
                    HAS_TO_PLOT_SAVED : bool (False)

                      If True, some curves can be plotted during the run.

                    kzkh_periodicity : int (0)

                      Periodicity of saving of (kz, kh) spectra (compared to standard spectra).
        """
            )
        )

    def _init_movies(self):
        self.movies = MoviesSpectra(self.output, self)

    def __init__(self, output):
        self.output = output

        if hasattr(self, "_init_movies"):
            self._init_movies()

        params = output.sim.params
        self.kzkh_periodicity = params.output.spectra.kzkh_periodicity
        self.nx = int(params.oper.nx)
        self.nb_saved_times = 0

        super().__init__(
            output,
            period_save=params.output.periods_save.spectra,
            has_to_plot_saved=params.output.spectra.HAS_TO_PLOT_SAVED,
        )

    def _init_path_files(self):
        path_run = self.output.path_run
        self.path_file1d = path_run + "/spectra1d.h5"
        self.path_file3d = path_run + "/spectra3d.h5"
        self.path_file_kzkh = path_run + "/spectra_kzkh.h5"

    def _init_files(self, arrays_1st_time=None):
        dict_spectra1d, dict_spectra3d, dict_kzkh = self.compute()
        if mpi.rank == 0:
            if not os.path.exists(self.path_file1d):
                oper = self.sim.oper
                kx = oper.deltakx * np.arange(oper.nkx_spectra)
                ky = oper.deltaky * np.arange(oper.nky_spectra)
                kz = oper.deltakz * np.arange(oper.nkz_spectra)
                self._create_file_from_dict_arrays(
                    self.path_file1d,
                    dict_spectra1d,
                    {"kx": kx, "ky": ky, "kz": kz},
                )
                self._create_file_from_dict_arrays(
                    self.path_file3d,
                    dict_spectra3d,
                    {"k_spectra3d": self.sim.oper.k_spectra3d},
                )
                if self.kzkh_periodicity:
                    self._create_file_from_dict_arrays(
                        self.path_file_kzkh,
                        dict_kzkh,
                        {"kz": kz, "kh_spectra": self.sim.oper.kh_spectra},
                    )

                self.nb_saved_times = 1
            else:
                with h5py.File(self.path_file1d, "r") as file:
                    dset_times = file["times"]
                    self.nb_saved_times = dset_times.shape[0] + 1
                # save the spectra in the file spectra1d.h5
                self._add_dict_arrays_to_file(self.path_file1d, dict_spectra1d)
                # save the spectra in the file spectra3d.h5
                self._add_dict_arrays_to_file(self.path_file3d, dict_spectra3d)

                if self.has_to_save_kzkh(only_rank0=True):
                    self._add_dict_arrays_to_file(self.path_file_kzkh, dict_kzkh)

        self.t_last_save = self.sim.time_stepping.t

    def has_to_save_kzkh(self, only_rank0=False):
        if mpi.rank == 0:
            answer = (
                self.kzkh_periodicity
                and self.nb_saved_times % self.kzkh_periodicity == 0
            )
        else:
            answer = None

        if only_rank0 or mpi.nb_proc == 1:
            return answer

        return mpi.comm.bcast(answer, root=0)

    def _online_save(self):
        """Save the values at one time. """
        tsim = self.sim.time_stepping.t
        if self._has_to_online_save():
            self.t_last_save = tsim
            dict_spectra1d, dict_spectra3d, dict_kzkh = self.compute()
            if mpi.rank == 0:
                # save the spectra in the file spectra1D.h5
                self._add_dict_arrays_to_file(self.path_file1d, dict_spectra1d)
                # save the spectra in the file spectra2D.h5
                self._add_dict_arrays_to_file(self.path_file3d, dict_spectra3d)
                if self.has_to_save_kzkh(only_rank0=True):
                    self._add_dict_arrays_to_file(self.path_file_kzkh, dict_kzkh)
                self.nb_saved_times += 1
                if self.has_to_plot:
                    self._online_plot_saving(dict_spectra1d, dict_spectra3d)

                    if tsim - self.t_last_show >= self.period_show:
                        self.t_last_show = tsim
                        self.axe.get_figure().canvas.draw()

    def compute(self):
        """compute the values at one time."""
        raise NotImplementedError

    def _init_online_plot(self):
        if mpi.rank == 0:
            fig, axe = self.output.figure_axe(numfig=1_000_000)
            self.axe = axe
            axe.set_xlabel("$k$")
            axe.set_ylabel("$E(k)$")
            axe.set_title("spectra\n" + self.output.summary_simul)

    def _online_plot_saving(self, dict_spectra1d, dict_spectra3d):
        pass

    def _load_mean_file(self, path, tmin=None, tmax=None, key_to_load=None):
        results = {}
        with h5py.File(path, "r") as h5file:
            times = h5file["times"][...]
            nt = len(times)
            if tmin is None:
                imin_plot = 0
            else:
                imin_plot = np.argmin(abs(times - tmin))
            if tmax is None:
                imax_plot = nt - 1
            else:
                imax_plot = np.argmin(abs(times - tmax))

            tmin = times[imin_plot]
            tmax = times[imax_plot]

            print(
                "compute mean of spectra\n"
                f"tmin = {tmin:8.6g} ; tmax = {tmax:8.6g}\n"
                f"imin = {imin_plot:8d} ; imax = {imax_plot:8d}"
            )

            if key_to_load is not None:
                if isinstance(key_to_load, str):
                    keys = [key_to_load]
                else:
                    keys = key_to_load
                for key in keys:
                    if key not in h5file.keys():
                        raise ValueError(f"{key} not in {h5file.keys()}")
                    spect = h5file[key][imin_plot : imax_plot + 1].mean(0)
                    results[key] = spect
                return results

            for key in list(h5file.keys()):
                if key.startswith("spectr"):
                    dset_key = h5file[key]
                    spect = dset_key[imin_plot : imax_plot + 1].mean(0)
                    results[key] = spect
        return results

    def load3d_mean(self, tmin=None, tmax=None):
        results = self._load_mean_file(self.path_file3d, tmin, tmax)
        with h5py.File(self.path_file3d, "r") as h5file:
            results["k"] = h5file["k_spectra3d"][...]
        return results

    def load1d_mean(self, tmin=None, tmax=None):
        results = self._load_mean_file(self.path_file1d, tmin, tmax)
        with h5py.File(self.path_file1d, "r") as h5file:
            for key in ("kx", "ky", "kz"):
                results[key] = h5file[key][...]
        return results

    def load_kzkh_mean(self, tmin=None, tmax=None, key_to_load=None):

        if not os.path.exists(self.path_file_kzkh):
            raise RuntimeError(
                self.path_file_kzkh
                + " does not exist. Can't load values from it."
            )

        results = self._load_mean_file(
            self.path_file_kzkh, tmin, tmax, key_to_load
        )
        with h5py.File(self.path_file_kzkh, "r") as h5file:
            for key in ("kz", "kh_spectra"):
                results[key] = h5file[key][...]

        return results

    def plot1d(self):
        pass

    def plot3d(self):
        pass



"""Main script for fluidsim command (:mod:`fluidsim.util.console.__main__`)
===========================================================================

"""
import argparse
from fluidsim import __version__, get_local_version
from . import bench, bench_analysis, profile
from .util import ConsoleError


def compute_name_command(module):
    name = module.__name__.split(".")[-1]
    return name.replace("_", "-")


def add_subparser(subparsers, module, description=None, subcommand=None):
    """Add a subparser for a module. Expects two functions `init_parser` and
    `run` to be defined within the module.

    """
    if subcommand is None:
        subcommand = compute_name_command(module)

    parser_subcommand = subparsers.add_parser(subcommand, description=description)
    try:
        module.init_parser(parser_subcommand)
    except ConsoleError:
        return

    parser_subcommand.set_defaults(func=module.run)


def print_verbose_version(args):
    """Prints verbose version including local SCM version if it exists."""
    print("fluidsim", __version__)
    local_version = get_local_version()
    if local_version != __version__:
        print("local version:", get_local_version())


def get_parser():
    """Defines parser for command `fluidsim` and all its sub-commands."""
    parser = argparse.ArgumentParser(
        prog="fluidsim",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description="Console utilities for FluidSim",
    )
    parser.add_argument(
        "-V", "--version", action="version", version=f"%(prog)s {__version__}"
    )

    subparsers = parser.add_subparsers(
        help='see "fluidsim {subcommand} -h" for more details'
    )
    for module in (bench, bench_analysis, profile):
        add_subparser(subparsers, module, module.description)

    parser_version = subparsers.add_parser(
        "version", description=print_verbose_version.__doc__
    )
    parser_version.set_defaults(func=print_verbose_version)

    return parser


def run():
    """Parse arguments and execute the current script."""
    parser = get_parser()

    def print_help(args):
        """To print help when no subcommands are given."""
        parser.print_help()

    parser.set_defaults(func=print_help)
    args = parser.parse_args()
    args.func(args)


def _run_from_module(module):
    parser = argparse.ArgumentParser(
        prog="fluidsim" + compute_name_command(module),
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description=module.description,
    )

    module.init_parser(parser)
    args = parser.parse_args()
    module.run(args)


def run_profile():
    _run_from_module(profile)


def run_bench():
    _run_from_module(bench)


def run_bench_analysis():
    _run_from_module(bench_analysis)


if __name__ == "__main__":
    run()



#!/usr/bin/env python
# coding=utf8

import numpy as np
import glob
import os

import matplotlib.pylab as plt

from solveq2d import solveq2d


dir_base = "Approach_runs_2048x2048"
set_of_dir_results = solveq2d.SetOfDirResults(dir_base=dir_base)

values_c2 = set_of_dir_results.values_c2
values_c2 = [200]
values_solver = set_of_dir_results.values_solver
# values_solver = ['SW1l']

tmin = 263


tuple_loop = [
    (c2, name_solver) for c2 in values_c2 for name_solver in values_solver
]
for c2, name_solver in tuple_loop:
    path_dir = set_of_dir_results.one_path_from_values(solver=name_solver, c2=c2)

    sim = solveq2d.load_state_phys_file(t_approx=1000, name_dir=path_dir)

    sim.output.spect_energy_budg.plot(tmin=tmin, tmax=500, delta_t=0.0)
    sim.output.spectra.plot2D(
        tmin=tmin, tmax=500, delta_t=0.0, coef_compensate=5.0 / 3
    )

    sim.output.spatial_means.plot()
    # sim.output.prob_dens_func.plot(tmin=tmin)

    # sim.output.phys_fields.plot(key_field='rot')
    # sim.output.phys_fields.plot(key_field='eta')


plt.show()



"""SEB (:mod:`fluidsim.solvers.ns3d.strat.output.spect_energy_budget`)
======================================================================

.. autoclass:: SpectralEnergyBudgetNS3D
   :members:
   :private-members:

"""

from textwrap import dedent

import numpy as np
import h5py

from fluidfft.fft3d.operators import vector_product

from fluidsim.base.output.base import SpecificOutput


class SpectralEnergyBudgetNS3D(SpecificOutput):
    r"""Spectral energy budget of ns3d.strat.

    .. |vv| mathmacro:: \textbf{v}

    .. |kk| mathmacro:: \textbf{k}

    .. |convK2A| mathmacro:: C_{K\rightarrow A}

    .. |bnabla| mathmacro:: \boldsymbol{\nabla}

    .. |bomega| mathmacro:: \boldsymbol{\omega}

    Notes
    -----

    .. math::

      d_t E_K(\kk) = T_K(\kk) - D_K(\kk),

    where :math:`E_K(\kk) = |\hat{\vv}|^2/2`.

    The transfer term is

    .. math::

      T_K(\kk) = \Re \left(\hat{\vv}^* \cdot P_\perp\widehat{\vv \times \bomega} \right).

    By definition, we have to have :math:`\sum T_K(\kk) = 0`.

    The dissipative term is

    .. math::

      D_K(\kk) = f_{dK}(\kk) E_K(\kk),

    where :math:`f_{dK}(\kk)` is the dissipation frequency depending of the
    wavenumber and the viscous coefficients.

    """

    _tag = "spect_energy_budg"
    _name_file = _tag + ".h5"

    @classmethod
    def _complete_params_with_default(cls, params):
        tag = cls._tag

        params.output.periods_save._set_attrib(tag, 0)
        p_seb = params.output._set_child(
            tag, attribs={"HAS_TO_PLOT_SAVED": False}
        )
        p_seb._set_doc(
            dedent(
                """
                    HAS_TO_PLOT_SAVED : bool (False)

                      If True, some curves can be plotted during the run.
        """
            )
        )

    def __init__(self, output):

        params = output.sim.params
        self.nx = params.oper.nx

        self.oper = oper = output.sim.oper

        kx = oper.deltakx * np.arange(oper.nkx_spectra)
        ky = oper.deltaky * np.arange(oper.nky_spectra)
        kz = oper.deltakz * np.arange(oper.nkz_spectra)

        super().__init__(
            output,
            period_save=params.output.periods_save.spect_energy_budg,
            has_to_plot_saved=params.output.spect_energy_budg.HAS_TO_PLOT_SAVED,
            arrays_1st_time={"kx": kx, "ky": ky, "kz": kz, "kh": oper.kh_spectra},
        )

    def compute_spectra(self, name, quantity):
        spectrum_kzkh = self.oper.compute_spectrum_kzkh(quantity)
        spectrum_kx, spectrum_ky, spectrum_kz = self.oper.compute_1dspectra(
            quantity
        )
        return {
            name: spectrum_kzkh,
            name + "_kx": spectrum_kx,
            name + "_ky": spectrum_ky,
            name + "_kz": spectrum_kz,
        }

    def compute(self):

        results = {}
        state = self.sim.state

        oper = self.sim.oper
        ifft_as_arg_destroy = oper.ifft_as_arg_destroy

        state_spect = state.state_spect

        vx_fft = state_spect.get_var("vx_fft")
        vy_fft = state_spect.get_var("vy_fft")
        vz_fft = state_spect.get_var("vz_fft")

        omegax_fft, omegay_fft, omegaz_fft = oper.rotfft_from_vecfft(
            vx_fft, vy_fft, vz_fft
        )

        omegax = state.fields_tmp[3]
        omegay = state.fields_tmp[4]
        omegaz = state.fields_tmp[5]

        ifft_as_arg_destroy(omegax_fft, omegax)
        ifft_as_arg_destroy(omegay_fft, omegay)
        ifft_as_arg_destroy(omegaz_fft, omegaz)

        del omegax_fft, omegay_fft, omegaz_fft

        state_phys = state.state_phys
        vx = state_phys.get_var("vx")
        vy = state_phys.get_var("vy")
        vz = state_phys.get_var("vz")

        fx, fy, fz = vector_product(vx, vy, vz, omegax, omegay, omegaz)

        fx_fft = oper.fft(fx)
        fy_fft = oper.fft(fy)
        fz_fft = oper.fft(fz)

        del fx, fy, fz

        oper.project_perpk3d(fx_fft, fy_fft, fz_fft)

        results.update(
            self.compute_spectra(
                "transfer_Kh",
                np.real(vx_fft.conj() * fx_fft) + np.real(vy_fft.conj() * fy_fft),
            )
        )

        results.update(
            self.compute_spectra("transfer_Kz", np.real(vz_fft.conj() * fz_fft))
        )

        urx_fft, ury_fft, _, _ = self.sim.oper.urudfft_from_vxvyfft(
            vx_fft, vy_fft
        )

        results.update(
            self.compute_spectra(
                "transfer_Khr",
                np.real(urx_fft.conj() * fx_fft)
                + np.real(ury_fft.conj() * fy_fft),
            )
        )

        del urx_fft, ury_fft, _

        f_d, f_d_hypo = self.sim.compute_freq_diss()
        del f_d_hypo

        results.update(
            self.compute_spectra(
                "diss_Kh", f_d * (abs(vx_fft) ** 2 + abs(vy_fft) ** 2)
            )
        )

        del vx_fft, vy_fft

        results.update(self.compute_spectra("diss_Kz", f_d * abs(vz_fft) ** 2))

        transfer_K = results["transfer_Kh"] + results["transfer_Kz"]
        assert transfer_K.sum() < 1e-14

        return results

    def load_mean(self, tmin=0, tmax=None, keys_to_load=None):
        means = {}
        with h5py.File(self.path_file, "r") as file:
            times = file["times"][...]
            nt = len(times)
            if tmin is None:
                imin_plot = 0
            else:
                imin_plot = np.argmin(abs(times - tmin))
            if tmax is None:
                imax_plot = nt - 1
            else:
                imax_plot = np.argmin(abs(times - tmax))

            tmin = times[imin_plot]
            tmax = times[imax_plot]

            print(
                "compute mean spectral energy budget\n"
                f"tmin = {tmin:8.6g} ; tmax = {tmax:8.6g}\n"
                f"imin = {imin_plot:8d} ; imax = {imax_plot:8d}"
            )

            for key in list(file.keys()):
                if key.startswith("k"):
                    means[key] = file[key][...]

            if keys_to_load is not None:
                if isinstance(keys_to_load, str):
                    keys_to_load = [keys_to_load]
                for key in keys_to_load:
                    if key not in file.keys():
                        print(key, file.keys())
                        raise ValueError
            else:
                keys_to_load = [
                    key
                    for key in file.keys()
                    if key != "times" and not key.startswith(("k", "info"))
                ]

            for key in keys_to_load:
                spect = file[key][imin_plot : imax_plot + 1].mean(0)
                means[key] = spect
        return means

    _key_plot_default_kzkh = "transfer_Kh"

    def plot_kzkh(self, tmin=0, tmax=None, key=None, ax=None):

        with h5py.File(self.path_file, "r") as file:
            keys_saved = [
                key
                for key in file.keys()
                if key not in ("times", "info_simul")
                and not key.startswith("k")
                and not any(key.endswith("_k" + letter) for letter in "xyz")
            ]

        if key is None:
            key = self._key_plot_default_kzkh

        if key not in keys_saved:
            raise ValueError(f"key '{key}' not in {keys_saved}")

        data = self.load_mean(tmin, tmax, key)
        spectrum = data[key]
        kz = data["kz"]
        kh = data["kh"]

        if ax is None:
            fig, ax = self.output.figure_axe()

        ax.set_xlabel(r"$\kappa_h$")
        ax.set_ylabel("$k_z$")
        ax.set_title(f"{key}\n{self.output.summary_simul}")

        ax.pcolormesh(kh, kz, spectrum, shading="nearest")

    def compute_fluxes_mean(self, tmin=None, tmax=None):

        with h5py.File(self.path_file, "r") as file:
            keys_saved = [
                key
                for key in file.keys()
                if key not in ("times", "info_simul")
                and not key.startswith("k")
                and not any(key.endswith("_k" + letter) for letter in "xyz")
            ]

        data = self.load_mean(tmin, tmax, keys_saved)

        kz = data["kz"]
        kh = data["kh"]

        dict_results = {"kz": kz, "kh": kh}

        deltakz = kz[1]
        deltakh = kh[1]

        for key in keys_saved:
            spectrum = data[key]
            flux = deltakz * spectrum.sum(0)
            flux = deltakh * np.cumsum(flux)
            if key.startswith("transfer"):
                flux *= -1

            key_flux = "hflux_" + key
            dict_results.update({key_flux: flux})

            flux = deltakh * spectrum.sum(1)
            flux = deltakz * np.cumsum(flux)
            if key.startswith("transfer"):
                flux *= -1

            key_flux = "zflux_" + key
            dict_results.update({key_flux: flux})

        return dict_results

    def plot_fluxes(self, tmin=None, tmax=None, key_k="kh", ax=None):

        data = self.compute_fluxes_mean(tmin, tmax)

        k_plot = data[key_k]
        deltak = k_plot[1]
        k_plot += deltak / 2

        key_flux = key_k[1] + "flux_"
        flux_Kh = data[key_flux + "transfer_Kh"]
        flux_Kz = data[key_flux + "transfer_Kz"]
        DKh = data[key_flux + "diss_Kh"]
        DKz = data[key_flux + "diss_Kz"]

        flux_tot = flux_Kh + flux_Kz
        D = DKh + DKz

        eps = D[-1]

        if ax is None:
            fig, ax = self.output.figure_axe()

        xlbl = "k_" + key_k[1]
        ylbl = r"$\Pi(" + xlbl + r")/\epsilon$"
        xlbl = "$" + xlbl + "$"
        ax.set_xlabel(xlbl)
        ax.set_ylabel(ylbl)
        ax.set_title(f"spectral fluxes\n{self.output.summary_simul}")

        ax.semilogx(
            k_plot, flux_tot / eps, "k", linewidth=2, label=r"$\Pi/\epsilon$"
        )
        ax.semilogx(k_plot, D / eps, "k--", linewidth=2, label=r"$D/\epsilon$")
        ax.semilogx(
            k_plot, (flux_tot + D) / eps, "k:", label=r"$(\Pi+D)/\epsilon$"
        )

        ax.legend()



"""Forcing (:mod:`fluidsim.solvers.ns2d.forcing`)
=================================================

.. autoclass:: ForcingNS2D
   :members:

"""

from fluidsim.base.forcing import ForcingBasePseudoSpectral
from fluidsim.base.forcing.anisotropic import (
    TimeCorrelatedRandomPseudoSpectralAnisotropic,
)
from fluidsim.base.forcing.milestone import ForcingMilestone


class ForcingNS2D(ForcingBasePseudoSpectral):
    """Forcing class for the ns2d solver.

    .. inheritance-diagram:: ForcingNS2D

    """

    @staticmethod
    def _complete_info_solver(info_solver):
        """Complete the ParamContainer info_solver."""
        classes = [
            TimeCorrelatedRandomPseudoSpectralAnisotropic,
            ForcingMilestone,
        ]
        ForcingBasePseudoSpectral._complete_info_solver(info_solver, classes)

    def compute_coef_ab_normalize(
        self, constant_rate_of, key_forced, f_fft, var_fft, deltat
    ):

        if constant_rate_of not in ["energy", "energyK"]:
            raise ValueError

        if hasattr(self.forcing_maker, "oper_coarse"):
            oper = self.forcing_maker.oper_coarse
        else:
            oper = self.sim.oper

        if key_forced == "rot_fft":
            vx_fft, vy_fft = oper.vecfft_from_rotfft(var_fft)
            fx_fft, fy_fft = oper.vecfft_from_rotfft(f_fft)
        else:
            raise ValueError

        a = deltat / 2 * oper.sum_wavenumbers(abs(fx_fft) ** 2 + abs(fy_fft) ** 2)
        b = oper.sum_wavenumbers(
            (vx_fft.conj() * fx_fft).real + (vy_fft.conj() * fy_fft).real
        )

        return a, b



"""Burgers1D solver (:mod:`fluidsim.solvers.burgers1d.skew_sym.solver`)
=======================================================================

Provides:

.. autoclass:: Simul
   :members:
   :private-members:


"""
from fluidsim.base.setofvariables import SetOfVariables
from ..solver import Simul as SimulBurgers, InfoSolver as InfoSolverBurgers


class InfoSolver(InfoSolverBurgers):
    def _init_root(self):
        super()._init_root()

        package = "fluidsim.solvers.burgers1d.skew_sym"
        self.module_name = package + ".solver"
        self.class_name = "Simul"
        self.short_name = "BurgersSkewSym1D"


class Simul(SimulBurgers):
    InfoSolver = InfoSolver

    def tendencies_nonlin(self, state_spect=None, old=None):

        if state_spect is None:
            u_fft = self.state.state_spect.get_var("u_fft")
            signal = self.state.state_phys.get_var("u")
        else:
            u_fft = state_spect.get_var("u_fft")
            signal = self.oper.ifft(u_fft)

        pxu_fft = self.oper.pxffft_from_fft(u_fft)
        pxu = self.oper.ifft(pxu_fft)

        # Half from convective form (in physical space)
        f_signal = -0.5 * signal * pxu

        # Half from conservative form (in spectral space)
        f_fft_cons = -0.25 * self.oper.pxffft_from_fft(self.oper.fft(signal ** 2))

        if old is None:
            tendencies_fft = SetOfVariables(
                like=self.state.state_spect, info="tendencies_nonlin"
            )
        else:
            tendencies_fft = old

        f_fft = tendencies_fft.get_var("u_fft")
        self.oper.fft_as_arg(f_signal, f_fft)

        # Add the conservative form part to yield the skew symmetric form
        f_fft[:] += f_fft_cons

        self.oper.dealiasing(f_fft)
        # Set "oddball mode" to zero
        f_fft[self.oper.nkx-1] = 0.
        return tendencies_fft



import unittest
import warnings

try:
    import scipy.sparse

    scipy_installed = True
except ImportError:
    scipy_installed = False


from fluiddyn.util import mpi

from fluidsim.solvers.ad1d.solver import Simul

from fluidsim.util.testing import TestSimul


@unittest.skipIf(not scipy_installed, "No module named scipy.sparse")
class TestSolverAD1D(TestSimul):

    Simul = Simul

    @classmethod
    def setUpClass(cls):
        super().setUpClass()

        # RuntimeWarnings are very common when numpy installed and numpy
        # used to build scipy don't match.
        # See:
        # [1] http://thread.gmane.org/gmane.comp.python.cython.devel/14352/focus=14354  # noqa
        # [2] https://stackoverflow.com/a/40846742
        warnings.filterwarnings(
            "ignore", "^numpy.ufunc size changed", RuntimeWarning
        )
        warnings.filterwarnings(
            "ignore", "^numpy.dtype size changed", RuntimeWarning
        )

    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()
        warnings.resetwarnings()

    @classmethod
    def init_params(cls):
        params = cls.params = cls.Simul.create_default_params()

        params.U = 1.0

        params.short_name_type_run = "test"
        params.output.sub_directory = "unittests"

        params.oper.nx = 40
        params.oper.Lx = 1.0

        params.time_stepping.type_time_scheme = "RK2"

        params.nu_2 = 0.01

        params.time_stepping.t_end = 0.4
        params.time_stepping.USE_CFL = True

        params.init_fields.type = "gaussian"

        params.output.periods_print.print_stdout = 0.25
        params.output.periods_save.phys_fields = 0.5
        params.output.periods_plot.phys_fields = 0.0
        params.output.phys_fields.field_to_plot = "s"

    @unittest.skipIf(
        mpi.nb_proc > 1, "MPI not implemented, for eg. sim.oper.gather_Xspace"
    )
    def test_simul(self):
        sim = self.sim
        sim.time_stepping.start()

        sim.state.compute("dx_s")
        dx_s = sim.state.compute("dx_s")

        with self.assertRaises(ValueError):
            sim.state.compute("bar")

        if hasattr(sim.oper, "identity"):
            sim.oper.identity()
            sim.oper.pxx(dx_s)

        sim.output.phys_fields.plot()
        sim.output.phys_fields.plot(field="s", time=10)


@unittest.skipIf(not scipy_installed, "No module named scipy.sparse")
class TestInitAD1D(TestSimul):
    Simul = Simul

    @classmethod
    def init_params(cls):
        params = cls.params = cls.Simul.create_default_params()
        params.output.HAS_TO_SAVE = False
        params.init_fields.type = "cos"

    @unittest.skipIf(
        mpi.nb_proc > 1, "MPI not implemented, for eg. sim.oper.gather_Xspace"
    )
    def test_init(self):
        """Only test the initialization"""


if __name__ == "__main__":
    unittest.main()



"""State for the AD1D solver (:mod:`fluidsim.solvers.ad1d.state`)
=======================================================================
"""

from fluidsim.base.state import StateBase

from fluiddyn.util import mpi


class StateAD1D(StateBase):
    """Contains the variables corresponding to the state and handles the
    access to other fields for the solver AD1D.

    """

    @staticmethod
    def _complete_info_solver(info_solver):
        """Complete the ParamContainer info_solver.

        This is a static method!
        """
        info_solver.classes.State._set_attribs(
            {
                "keys_state_phys": ["s"],
                "keys_computable": [],
                "keys_phys_needed": ["s"],
                "keys_linear_eigenmodes": ["s"],
            }
        )

    def compute(self, key, SAVE_IN_DICT=True):
        it = self.sim.time_stepping.it
        if key in self.vars_computed and it == self.it_computed[key]:
            return self.vars_computed[key]

        if key == "dx_s":
            result = self.oper.px(self.state_phys.get_var("s"))

        else:
            raise ValueError('Do not know how to compute "' + key + '".')

        if SAVE_IN_DICT:
            self.vars_computed[key] = result
            self.it_computed[key] = it

        return result



#!/usr/bin/env python
# coding=utf8

import numpy as np
import matplotlib.pyplot as plt

from solveq2d import solveq2d

from create_figs_articles import CreateFigArticles


num_fig = 1000
SAVE_FIG = 1
name_file = "spectra_forcingq_diff_c.eps"

create_fig = CreateFigArticles(
    short_name_article="SW1l", SAVE_FIG=SAVE_FIG, FOR_BEAMER=False, fontsize=19
)

dir_base = (
    create_fig.path_base_dir + "/Results_for_article_SW1l"
    "/Approach_runs_2048x2048"
)
set_of_dir_results = solveq2d.SetOfDirResults(dir_base=dir_base)
dirs = set_of_dir_results.dirs_from_values(solver="SW1lexlin", FORCING=True)

print(dirs)

tmin = 264


def sprectra_from_namedir(name_dir_results):
    path_dir_results = set_of_dir_results.path_dirs[name_dir_results]
    sim = solveq2d.create_sim_plot_from_dir(path_dir_results)
    dict_results = sim.output.spectra.load2D_mean(tmin=tmin)
    kh = dict_results["kh"]
    EK = dict_results["spectrum2D_EK"]
    EA = dict_results["spectrum2D_EA"]
    EKr = dict_results["spectrum2D_EKr"]
    EKd = EK - EKr
    return kh, EKr, EKd


kh, EKr0, EKd0 = sprectra_from_namedir(dirs[0])
kh, EKr1, EKd1 = sprectra_from_namedir(dirs[1])
kh, EKr2, EKd2 = sprectra_from_namedir(dirs[2])
kh, EKr3, EKd3 = sprectra_from_namedir(dirs[3])


fig, ax1 = create_fig.figure_axe(name_file=name_file)
ax1.set_xscale("log")
ax1.set_yscale("log")

coef_compensate = 5.0 / 3
coef_norm = kh ** coef_compensate

ax1.plot(kh, EKr0 * coef_norm, "k--", linewidth=2)
ax1.plot(kh, EKd0 * coef_norm, "k:", linewidth=2)

ax1.plot(kh, EKr1 * coef_norm, "r--", linewidth=2)
ax1.plot(kh, EKd1 * coef_norm, "r:", linewidth=2)

ax1.plot(kh, EKr2 * coef_norm, "b--", linewidth=2)
ax1.plot(kh, EKd2 * coef_norm, "b:", linewidth=2)

ax1.plot(kh, EKr3 * coef_norm, "c--", linewidth=2)
ax1.plot(kh, EKd3 * coef_norm, "c:", linewidth=2)


cond = np.logical_and(kh > 1, kh < 20)
ax1.plot(kh[cond], 1e1 * kh[cond] ** (-3.0) * coef_norm[cond], "k--", linewidth=1)
plt.figtext(0.6, 0.78, "$k^{-3}$", fontsize=20)

cond = np.logical_and(kh > 0.3, kh < 10)
ax1.plot(kh[cond], 4e-2 * kh[cond] ** (-2.0) * coef_norm[cond], "k:", linewidth=1)
plt.figtext(0.25, 0.55, "$k^{-2}$", fontsize=20)

cond = np.logical_and(kh > 0.3, kh < 15)
ax1.plot(
    kh[cond], 1e-2 * kh[cond] ** (-5.0 / 3) * coef_norm[cond], "k-.", linewidth=1
)
plt.figtext(0.5, 0.45, "$k^{-5/3}$", fontsize=20)


ax1.set_xlabel("$k_h$")
ax1.set_ylabel("2D spectra")


# plt.rc('legend', numpoints=1)
# leg1 = plt.figlegend(
#         [l_Etot[0], l_EK[0], l_EA[0]],
#         ['$E$', '$E_K$', '$E_A$'],
#         loc=(0.78, 0.7),
#         labelspacing = 0.2
# )


ax1.set_xlim([0.1, 150])
ax1.set_ylim([1e-5, 2e1])


create_fig.save_fig()

plt.show()



"""
find_dissipation (:mod:`fluidsim.solvers.ns2d.strat.find_dissipation`)
==========================================================================

To execute:

python fluidsim/solvers/ns2d/strat/find_dissipation.py 240 0.5 "nu_8"
"""

import os
import h5py
import shutil
import numpy as np
import matplotlib.pyplot as plt

from pathlib import Path
from math import pi
from glob import glob
from copy import deepcopy as _deepcopy

from fluidsim import load_state_phys_file
from fluidsim.solvers.ns2d.strat.solver import Simul
from fluiddyn.util import mpi


def make_parameters_simulation(gamma, key_viscosity):
    """ Make parameters of the first simulation. """
    # Parameters simulation
    F = np.sin(pi / 4)
    sigma = 1

    # Create parameters
    params = Simul.create_default_params()

    # Operator parameters
    params.oper.nx = nx
    params.oper.ny = nz = nx // 4
    params.oper.Lx = Lx = 2 * pi
    params.oper.Ly = Lz = Lx * (nz / nx)
    params.oper.NO_SHEAR_MODES = True
    params.oper.coef_dealiasing = 0.6666

    # Forcing parameters
    params.forcing.enable = True
    params.forcing.type = "tcrandom_anisotropic"
    params.forcing.key_forced = "ap_fft"
    params.forcing.nkmax_forcing = nkmax_forcing = 8
    params.forcing.nkmin_forcing = nkmin_forcing = 4

    params.forcing.normalized.constant_rate_of = "energy"

    params.forcing.tcrandom_anisotropic.angle = np.arcsin(F)

    # Compute other parameters (Normalization by the energy..)
    tau_af = 1  # Forcing time equal to 1
    k_f = ((nkmax_forcing + nkmin_forcing) / 2) * max(2 * pi / Lx, 2 * pi / Lz)
    forcing_rate = (1 / tau_af ** 3) * ((2 * pi) / k_f) ** 2
    omega_af = 2 * pi / tau_af

    params.N = (gamma / F) * omega_af

    # Choose vis, key_viscosity=key_viscositycosity
    if key_viscosity == "nu_2":
        params.nu_2 = nu_2
    elif key_viscosity == "nu_8":
        params.nu_8 = nu_8
    else:
        raise ValueError(f"{key_viscosity} not known.")

    # Continuation on forcing...
    params.forcing.forcing_rate = forcing_rate
    params.forcing.tcrandom.time_correlation = sigma * (
        pi / (params.N * F)
    )  # time_correlation = wave period

    # Time stepping parameters
    params.time_stepping.USE_CFL = True
    params.time_stepping.t_end = t_end

    # Initialization parameters
    params.init_fields.type = "noise"

    modify_parameters(params)

    return params


def modify_parameters(params):
    """ Function modifies default parameters. """
    # Output parameters
    params.output.HAS_TO_SAVE = True
    params.output.sub_directory = "find_diss_coef"
    params.output.periods_save.phys_fields = 0.4
    params.output.periods_save.spatial_means = 1e-1
    params.output.periods_save.spect_energy_budg = 1e-1
    params.output.periods_save.spectra = 1e-1


def compute_diff(idx_dealiasing, idy_dealiasing, dissE_kx, dissE_ky):
    """Computes the number of modes between the dissipation peak and the largest modes. """
    idx_diss_max = np.argmax(abs(dissE_kx))
    idy_diss_max = np.argmax(abs(dissE_ky))

    # Computes difference
    diff_x = idx_dealiasing - idx_diss_max
    diff_y = idy_dealiasing - idy_diss_max

    diff = np.argmax([diff_x, diff_y])

    if diff == 0:
        print(
            "diff_x = idx_dealiasing -  idx_diss_max",
            idx_dealiasing - idx_diss_max,
        )
        diff = idx_dealiasing - idx_diss_max

    else:
        print(
            "diff_y = idy_dealiasing -  idy_diss_max",
            idy_dealiasing - idy_diss_max,
        )
        diff = idy_dealiasing - idy_diss_max

    return diff, idx_diss_max, idy_diss_max


def normalization_initialized_field(sim, coef_norm=1e-4):
    """Normalizes the initialized field. (ONLY if nx != ny)"""
    if sim.params.oper.nx != sim.params.oper.ny:

        if not sim.params.forcing.key_forced == "ap_fft":
            raise ValueError("sim.params.forcing.key_forced should be ap_fft.")

        KX = sim.oper.KX
        cond = KX == 0.0

        ux_fft = sim.state.get_var("ux_fft")
        uy_fft = sim.state.get_var("uy_fft")
        b_fft = sim.state.get_var("b_fft")

        ux_fft[cond] = 0.0
        uy_fft[cond] = 0.0
        b_fft[cond] = 0.0

        # Compute energy after ux_fft[kx=0] uy_fft[kx=0] b_fft[kx=0]
        ek_fft = (np.abs(ux_fft) ** 2 + np.abs(uy_fft) ** 2) / 2
        ea_fft = ((np.abs(b_fft) / params.N) ** 2) / 2
        e_fft = ek_fft + ea_fft
        energy_before_norm = sim.output.sum_wavenumbers(e_fft)

        # Compute scale energy forcing
        Lx = sim.params.oper.Lx
        Lz = sim.params.oper.Ly

        nkmax_forcing = params.forcing.nkmax_forcing
        nkmin_forcing = params.forcing.nkmin_forcing

        k_f = ((nkmax_forcing + nkmin_forcing) / 2) * max(
            2 * pi / Lx, 2 * pi / Lz
        )
        energy_f = params.forcing.forcing_rate ** (2 / 3) * (2 * pi / k_f) ** (
            2 / 3
        )

        coef = np.sqrt(coef_norm * energy_f / energy_before_norm)

        ux_fft *= coef
        uy_fft *= coef
        b_fft *= coef

        rot_fft = sim.oper.rotfft_from_vecfft(ux_fft, uy_fft)
        sim.state.init_statespect_from(rot_fft=rot_fft, b_fft=b_fft)

        sim.state.statephys_from_statespect()
    else:
        pass
    return sim


def _compute_ikdiss(dissE_kx, dissE_ky):
    """ Computes index peak dissipation. """
    idx_diss_max = np.argmax(abs(dissE_kx))
    idy_diss_max = np.argmax(abs(dissE_ky))

    return idx_diss_max, idy_diss_max


def _compute_ikmax(kxE, kyE):
    """ Computes index largest modes. """
    idx_dealiasing = np.argmin(abs(kxE - sim.oper.kxmax_dealiasing))
    idy_dealiasing = np.argmin(abs(kyE - sim.oper.kymax_dealiasing))

    return idx_dealiasing, idy_dealiasing


def compute_delta_ik(kxE, kyE, dissE_kx, dissE_ky):
    """ Computes the difference between the dissipation peak and target. """
    idx_diss_max, idy_diss_max = _compute_ikdiss(dissE_kx, dissE_ky)
    idx_dealiasing, idy_dealiasing = _compute_ikmax(kxE, kyE)

    idx_target = idx_dealiasing - (nb_wavenumbers_x - 1)
    idy_target = idy_dealiasing - (nb_wavenumbers_y - 1)

    delta_ikx = idx_target - idx_diss_max
    delta_iky = idy_target - idy_diss_max

    return idx_target, idy_target, delta_ikx, delta_iky


def compute_energy_spatial(sim):
    """ Compute energy without energy in shear modes """
    dict_spatial = sim.output.spatial_means.load()
    E = dict_spatial["E"] - dict_spatial["E_shear"]
    t = dict_spatial["t"]

    PK_tot = dict_spatial["PK_tot"]
    PA_tot = dict_spatial["PA_tot"]
    P_tot = PK_tot + PA_tot
    return E, t, P_tot


def modify_factor(sim, key_viscosity):
    """ Modifies factor for viscosity. """
    params = _deepcopy(sim.params)

    params_old = sim.params
    sim_old = sim

    if key_viscosity == "nu_2":
        nu_old = params.nu_2
        params.nu_2 = params.nu_2 * factor
    elif key_viscosity == "nu_8":
        nu_old = params.nu_8
        params.nu_8 = params.nu_8 * factor
    else:
        raise ValueError

    params.init_fields.type = "in_script"
    params.time_stepping.t_end = t_end

    return params, nu_old


def write_to_file(path, to_print, mode="a"):
    """ Writes to a file."""
    with open(path, mode) as file:
        file.write(to_print)


def make_float_value_for_path(value):
    """ Makes float. """
    value_not = str(value)
    if "." in value_not:
        return value_not.split(".")[0] + "_" + value_not.split(".")[1]
    else:
        return value_not


def check_dissipation():
    """ Checks dissipation"""
    ratio_x = dissE_kx[idx_diss_max] / dissE_kx[idx_dealiasing - 1]
    ratio_y = dissE_ky[idy_diss_max] / dissE_ky[idy_dealiasing - 1]

    idx_target, idy_target, delta_ikx, delta_iky = compute_delta_ik(
        kxE, kyE, dissE_kx, dissE_ky
    )

    if time_total > 1000:
        print("The stationarity has not " + f"reached after {it} simulations.")
        should_I_stop = "non_stationarity"

    if ratio_x > threshold_ratio and ratio_y > threshold_ratio:

        if delta_ikx > 0 and delta_iky > 0:

            if delta_ikx > delta_iky:
                print("limited by y")
                norm = idy_dealiasing
                delta_ikmin = delta_iky + nb_wavenumbers_y // 4
            else:
                print("limited by x")
                norm = idx_dealiasing
                delta_ikmin = delta_ikx + nb_wavenumbers_x // 4

            factor = max((1 - (delta_ikmin / norm)) ** 1.5, 0.85)
            # factor = max((1 - (delta_ikmin / norm)) ** 1.5, 0.5)
            print("factor", factor)
            should_I_stop = False

        else:
            print(f"Checking stationarity... with {key_viscosity} = {nu_old}")
            E, t, P_tot = compute_energy_spatial(sim)
            ratio = np.mean(np.diff(E[2:]) / np.diff(t[2:]))

            if key_viscosity == "nu_2":
                condition_visc = abs(nu_old - params.nu_2) / params.nu_2 < 0.05
            elif key_viscosity == "nu_8":
                condition_visc = abs(nu_old - params.nu_8) / params.nu_8 < 0.05
            else:
                raise ValueError

            if (ratio / injection_energy_0) < 0.5 and condition_visc:
                if key_viscosity == "nu_2":
                    print(f"Stationarity is reached.\n nu = {params.nu_2}")
                elif key_viscosity == "nu_8":
                    print(f"Stationarity is reached.\n nu = {params.nu_8}")
                else:
                    raise ValueError

                factor = 1.0
                should_I_stop = True
            else:
                should_I_stop = False
                factor = 1.0
    else:

        factor = 1 + (1 / min(ratio_x, ratio_y))
        should_I_stop = False

    return factor, should_I_stop


if __name__ == "__main__":

    # SIMULATION
    nx = 240
    gamma = 0.5
    key_viscosity = "nu_8"
    nu_2 = 1e-3
    nu_8 = 1e-16
    t_end = 8.0
    PLOT_FIGURES = True

    # CONDITIONS
    threshold_ratio = 1e1
    min_factor = 0.7
    if nx == 240:
        nb_wavenumbers_y = 6
    elif nx == 480:
        nb_wavenumbers_y = 8
    elif nx == 960:
        nb_wavenumbers_y = 10
    else:
        raise ValueError

    gamma_not = make_float_value_for_path(gamma)

    # Create directory in path
    name_dir = "Coef_Diss_feb19"

    if key_viscosity == "nu_2":
        name_dir += "_nu2"
    elif key_viscosity == "nu_8":
        name_dir += "_nu8"
    else:
        raise ValueError

    path_root = Path("/fsnet/project/meige/2015/15DELDUCA/DataSim/")
    path_dir = path_root / name_dir
    directory_gamma = f"Coef_Diss_gamma{gamma_not}"
    path = path_dir / directory_gamma

    if mpi.rank == 0 and not path.exists():
        os.mkdir(path)

    # Check list simulations in directory
    paths_sim = sorted(path.glob("NS*"))

    # Write in .txt file
    path_file_write = path / "results.txt"

    PLOT_FIGURES = PLOT_FIGURES and mpi.rank == 0

    # Write temporal results in .txt file
    path_file_write2 = path / "results_check_nu.txt"
    if mpi.rank == 0:
        if path_file_write2.exists():
            os.remove(path_file_write2)

    if len(paths_sim) == 0:
        # Make FIRST SIMULATION
        params = make_parameters_simulation(gamma, key_viscosity)
        sim = Simul(params)
        sim = normalization_initialized_field(sim)
        sim.time_stepping.start()
    else:
        # Look for path largest resolution
        new_file = sorted(paths_sim[-1].glob("State_phys*"))[-1]
        path_file = sorted(new_file.glob("state_phys*"))[0].as_posix()

        # Compute resolution from path
        res_str = new_file.name.split("_")[-1]

        # sim = load_state_phys_file(paths_sim[-1])
        sim = load_state_phys_file(paths_sim[-1])
        params = _deepcopy(sim.params)

        params.oper.nx = int(res_str.split("x")[0])
        params.oper.ny = int(res_str.split("x")[1])

        params.init_fields.type = "from_file"
        params.init_fields.from_file.path = path_file

        params.time_stepping.t_end += t_end

        params.NEW_DIR_RESULTS = True

        modify_parameters(params)

        sim = Simul(params)
        sim.time_stepping.start()

    # Parameters condition
    nb_wavenumbers_x = nb_wavenumbers_y * (
        sim.params.oper.nx // sim.params.oper.ny
    )

    # Creation time and energy array
    time_total = 0
    time_total += sim.time_stepping.t

    energy, t, P_tot = compute_energy_spatial(sim)
    energy = np.mean(energy[len(energy) // 2 :])
    injection_energy_0 = P_tot[2]

    energies = []
    viscosities = []
    energies.append(energy)
    if key_viscosity == "nu_2":
        pnu = sim.params.nu_2
    elif key_viscosity == "nu_8":
        pnu = sim.params.nu_8
    else:
        raise ValueError
    viscosities.append(pnu)

    # Write results into temporal file
    if mpi.rank == 0:
        to_print = (
            "####\n"
            "t = {:.4e} \n"
            "E = {:.4e} \n"
            "nu2 = {:.4e} \n"
            "factor = {:.4e} \n"
        ).format(time_total, energy, pnu, 1)

        write_to_file(path_file_write2, to_print, mode="w")

    # Compute the data spectra energy budget
    kxE, kyE, dissE_kx, dissE_ky = sim.output.spect_energy_budg.load_mean(
        tmin=2.0, tmax=None
    )
    idx_diss_max, idy_diss_max = _compute_ikdiss(dissE_kx, dissE_ky)
    idx_dealiasing, idy_dealiasing = _compute_ikmax(kxE, kyE)
    idx_target, idy_target, delta_ikx, delta_iky = compute_delta_ik(
        kxE, kyE, dissE_kx, dissE_ky
    )

    # Plot dissipation
    if PLOT_FIGURES:
        plt.ion()
        fig, ax = plt.subplots()
        ax.set_title("$D(k_y)$")
        ax.set_xlabel("$k_y$")
        ax.set_ylabel("$D(k_y)$")
        ax.plot(
            kyE,
            dissE_ky,
            label="nu = {:.2e}, diff = {}".format(
                sim.params.nu_2, abs(idy_diss_max - idy_dealiasing)
            ),
        )
        ax.plot(sim.oper.kymax_dealiasing, 0, "xr")
        ax.axvline(x=sim.oper.deltaky * idy_target, color="k")

        fig2, ax2 = plt.subplots()
        ax2.set_title("$D(k_x)$")
        ax2.set_xlabel("$k_x$")
        ax2.set_ylabel("$D(k_x)$")
        ax2.plot(
            kxE,
            dissE_kx,
            label="nu = {:.2e}, diff = {}".format(
                sim.params.nu_2, abs(idx_diss_max - idx_dealiasing)
            ),
        )
        ax2.plot(sim.oper.kxmax_dealiasing, 0, "xr")
        ax2.axvline(x=sim.oper.deltakx * idx_target, color="k")

        ax.legend()
        ax2.legend()

        fig.canvas.draw()
        fig2.canvas.draw()

        plt.pause(1e-3)

        # Dissipation vs time
        fig3, ax3 = plt.subplots()
        ax3.set_xlabel("times")
        ax3.set_ylabel(r"$\nu_2$")

        ax3.plot(time_total, viscosities[-1], ".")

        # Energy Vs time
        fig4, ax4 = plt.subplots()
        ax4.plot(time_total, energy, ".")
        ax4.set_xlabel("times")
        ax4.set_ylabel("Energy")

        # Factor Vs time
        fig5, ax5 = plt.subplots()
        ax5.plot(time_total, 1, ".")
        ax5.set_xlabel("times")
        ax5.set_ylabel("Factor")

    it = 0
    p = 1
    # Check ...
    while True:

        if mpi.rank == 0:
            factor, should_I_stop = check_dissipation()
        else:
            factor = None
            should_I_stop = None

        if mpi.nb_proc > 1:
            # send factor and should_I_stop
            factor = mpi.comm.bcast(factor, root=0)
            should_I_stop = mpi.comm.bcast(should_I_stop, root=0)

        if should_I_stop:
            break

        if mpi.rank == 0:
            print("factor = ", factor)
            print("nu_2 OLD = ", sim.params.nu_2)
            print("nu_2 NEW = ", sim.params.nu_2 * factor)

        it += 1
        # Modification parameters
        params, nu_old = modify_factor(sim, key_viscosity)

        # Create new object simulation
        b_fft = sim.state.get_var("b_fft")
        rot_fft = sim.state.get_var("rot_fft")

        sim = Simul(params)
        sim.state.init_statespect_from(rot_fft=rot_fft, b_fft=b_fft)
        sim.state.statephys_from_statespect()
        sim.time_stepping.start()

        if mpi.rank == 0:
            # Add values to time array and energy array
            time_total += sim.time_stepping.t
            energy, t, P_tot = compute_energy_spatial(sim)
            energy = np.mean(energy[len(energy) // 2 :])
            energies.append(energy)
            viscosities.append(sim.params.nu_2)

            # Computes new index k_max_dissipation
            kxE, kyE, dissE_kx, dissE_ky = sim.output.spect_energy_budg.load_mean(
                tmin=2.0, tmax=None
            )
            idx_diss_max, idy_diss_max = _compute_ikdiss(dissE_kx, dissE_ky)
            idx_dealiasing, idy_dealiasing = _compute_ikmax(kxE, kyE)

            # Write results into temporal file
            to_print = (
                "####\n"
                "t = {:.4e} \n"
                "E = {:.4e} \n"
                "nu = {:.4e} \n"
                "factor = {:.4e} \n"
            ).format(time_total, energy, sim.params.nu_2, factor)

            write_to_file(path_file_write2, to_print, mode="a")

        if PLOT_FIGURES:
            ax.plot(
                kyE,
                dissE_ky,
                label="nu = {:.2e}, diff = {}".format(
                    params.nu_2, abs(idy_diss_max - idy_dealiasing)
                ),
            )
            ax2.plot(
                kxE,
                dissE_kx,
                label="nu = {:.2e}, diff = {}".format(
                    params.nu_2, abs(idx_diss_max - idx_dealiasing)
                ),
            )

            fig.canvas.draw()
            fig2.canvas.draw()

            plt.pause(1e-4)

            ax3.plot(time_total, viscosities[-1], "x")
            ax3.autoscale()
            fig3.canvas.draw()

            ax4.plot(time_total, energy, "x")
            ax4.autoscale()
            fig4.canvas.draw()

            ax5.plot(time_total, factor, "x")
            ax5.autoscale()
            fig5.canvas.draw()

            plt.pause(1e-4)

    if mpi.rank == 0:
        if len(paths_sim) == 0:
            to_print = "gamma,nx,nu \n"
            to_print += "{},{},{} \n".format(
                gamma, sim.params.oper.nx, params.nu_2
            )
            mode_write = "w"

        else:
            to_print = "{},{},{} \n".format(
                gamma, sim.params.oper.nx, params.nu_2
            )
            mode_write = "a"

        write_to_file(path_file_write, to_print, mode=mode_write)

        shutil.move(sim.params.path_run, path)



import unittest

import matplotlib.pyplot as plt

import fluidsim as fls

import fluiddyn.util.mpi as mpi

from fluidsim.solvers.ns2d.bouss.solver import Simul
from fluidsim.solvers.ns2d.test_solver import TestSimulBase as Base


class TestSimulBase(Base):
    Simul = Simul


class TestForcingOutput(TestSimulBase):
    @classmethod
    def init_params(self):

        params = super().init_params()
        params.forcing.enable = True
        params.forcing.type = "tcrandom"

        # save all outputs!
        periods = params.output.periods_save
        for key in periods._key_attribs:
            periods[key] = 0.2

        params.output.ONLINE_PLOT_OK = True
        params.output.periods_print.print_stdout = 0.2
        params.output.periods_plot.phys_fields = 0.2

        params.output.increments.HAS_TO_PLOT_SAVED = True

        for tag in params.output._tag_children:
            if tag.startswith("periods"):
                continue
            child = params.output[tag]
            if hasattr(child, "HAS_TO_PLOT_SAVED"):
                child["HAS_TO_PLOT_SAVED"] = True

    def test_forcing_output(self):

        sim = self.sim

        sim.time_stepping.start()

        sim.state.compute("rot_fft")
        sim.state.compute("rot_fft")

        sim.state.compute("ux_fft")
        sim.state.compute("uy_fft")

        if mpi.nb_proc == 1:
            plt.close("all")
            sim.output.spectra.plot1d()
            sim.output.spectra.plot2d()
            sim.output.spectra.load2d_mean()
            sim.output.spectra.load1d_mean()

            sim.output.spatial_means.plot()
            sim.output.spatial_means.plot_dt_energy()
            sim.output.spatial_means.plot_dt_enstrophy()
            sim.output.spatial_means.compute_time_means()
            sim.output.spatial_means.load_dataset()
            sim.output.spatial_means.time_first_saved()
            sim.output.spatial_means.time_last_saved()

            sim.output.print_stdout.plot_energy()
            sim.output.print_stdout.plot_deltat()

            sim.output.spectra_multidim.plot()

            sim.output.spect_energy_budg.plot()
            with self.assertRaises(ValueError):
                sim.state.get_var("test")

            sim2 = fls.load_sim_for_plot(sim.output.path_run)
            sim2.output

            sim2.output.increments.load()
            sim2.output.increments.plot()
            sim2.output.increments.load_pdf_from_file()
            sim2.output.increments.plot_pdf()

            sim2.output.phys_fields.animate(
                "ux",
                dt_frame_in_sec=1e-6,
                dt_equations=0.3,
                repeat=False,
                clim=(-1, 1),
                save_file=False,
                numfig=1,
            )
            sim2.output.phys_fields.plot()

            # `compute('q')` two times for better coverage...
            sim.state.get_var("q")
            sim.state.get_var("q")
            sim.state.get_var("div")

            path_run = sim.output.path_run
            if mpi.nb_proc > 1:
                path_run = mpi.comm.bcast(path_run)

            sim3 = fls.load_state_phys_file(path_run, modif_save_params=False)
            sim3.params.time_stepping.t_end += 0.2
            sim3.time_stepping.start()

        plt.close("all")


if __name__ == "__main__":
    unittest.main()



import unittest

import numpy as np
import matplotlib.pyplot as plt

import fluidsim as fls

import fluiddyn.util.mpi as mpi

from fluidsim.util.testing import TestSimul, classproperty, skip_if_no_fluidfft


@skip_if_no_fluidfft
class TestSimulBase(TestSimul):
    @classproperty
    def Simul(cls):
        from fluidsim.solvers.ns2d.solver import Simul

        return Simul

    @classmethod
    def init_params(cls):

        params = (
            cls.params
        ) = cls.Simul.create_default_params()  # pylint: disable=maybe-no-member
        params.short_name_type_run = "test"
        params.output.sub_directory = "unittests"

        nh = 32
        params.oper.nx = nh
        params.oper.ny = nh
        Lh = 6.0
        params.oper.Lx = Lh
        params.oper.Ly = Lh

        params.oper.coef_dealiasing = 2.0 / 3
        params.nu_8 = 2.0e-5

        params.time_stepping.t_end = 0.5

        params.init_fields.type = "noise"

        return params


class TestSolverNS2DTendency(TestSimulBase):
    @classmethod
    def init_params(self):
        params = super().init_params()

        params.output.HAS_TO_SAVE = False

    def test_tendency(self):
        sim = self.sim
        rot_fft = sim.state.get_var("rot_fft")

        tend = sim.tendencies_nonlin(state_spect=sim.state.state_spect)
        Frot_fft = tend.get_var("rot_fft")

        T_rot = np.real(Frot_fft.conj() * rot_fft)

        ratio = sim.oper.sum_wavenumbers(T_rot) / sim.oper.sum_wavenumbers(
            abs(T_rot)
        )

        self.assertGreater(1e-15, abs(ratio))


class TestForcingProportional(TestSimulBase):
    @classmethod
    def init_params(self):
        params = super().init_params()
        params.forcing.enable = True
        params.forcing.type = "proportional"

    def test_proportional(self):
        self.sim.time_stepping.start()
        self.sim.forcing.forcing_maker.verify_injection_rate()
        self.sim.forcing.forcing_maker.verify_injection_rate_coarse()


class TestForcing(TestSimulBase):
    @classmethod
    def init_params(self):
        params = super().init_params()
        params.forcing.enable = True
        params.forcing.type = "tcrandom"
        params.forcing.normalized.type = "particular_k"

    def test_(self):
        self.sim.time_stepping.start()


class TestForcingConstantRateEnergy(TestSimulBase):
    @classmethod
    def init_params(self):
        params = super().init_params()
        params.forcing.enable = True
        params.forcing.type = "tcrandom"
        params.forcing.normalized.constant_rate_of = "energy"
        params.forcing.forcing_rate = 3.333
        params.output.periods_save.spatial_means = 1e-6

    def test_(self):
        self.sim.time_stepping.start()

        if mpi.rank == 0:
            # Does the energy injection rate have the correct value at all times ?
            means = self.sim.output.spatial_means.load()
            assert np.allclose(
                means["PK_tot"], self.sim.params.forcing.forcing_rate
            )


class TestForcingOutput(TestSimulBase):
    @classmethod
    def init_params(self):

        params = super().init_params()

        params.oper.truncation_shape = "no_multiple_aliases"
        params.oper.NO_SHEAR_MODES = True
        params.oper.NO_KY0 = True

        params.forcing.enable = True
        params.forcing.type = "tcrandom"

        params.time_stepping.max_elapsed = "0:10:00"
        params.time_stepping.type_time_scheme = "RK2_phaseshift_random"

        params.nu_m4 = 1e-3

        # save all outputs!
        periods = params.output.periods_save
        for key in periods._key_attribs:
            periods[key] = 0.2

        params.output.ONLINE_PLOT_OK = True
        params.output.periods_print.print_stdout = 0.2
        params.output.periods_plot.phys_fields = 0.2

        params.output.increments.HAS_TO_PLOT_SAVED = True

        for tag in params.output._tag_children:
            if tag.startswith("periods"):
                continue
            child = params.output[tag]
            if hasattr(child, "HAS_TO_PLOT_SAVED"):
                child["HAS_TO_PLOT_SAVED"] = True

    def test_forcing_output(self):

        sim = self.sim
        assert f"{sim.params.oper.nx}x{sim.params.oper.ny}" in sim.name_run

        sim.time_stepping.start()

        # testing phaseshift
        phaseshift = sim.time_stepping._get_phaseshift()
        assert phaseshift.shape == sim.oper.KX.shape
        assert sim.time_stepping._get_phaseshift() is phaseshift

        sim.state.compute("rot_fft")
        sim.state.compute("rot_fft")

        with self.assertRaises(ValueError):
            sim.state.compute("abc")
        sim.state.compute("abc", RAISE_ERROR=False)

        ux_fft = sim.state.compute("ux_fft")
        uy_fft = sim.state.compute("uy_fft")

        sim.state.init_statespect_from(ux_fft=ux_fft)
        sim.state.init_statespect_from(uy_fft=uy_fft)

        # test_enstrophy_conservation
        # Verify that the enstrophy growth rate due to nonlinear tendencies
        # (advection term) must be zero.
        self.sim.params.forcing.enable = False
        tendencies_fft = self.sim.tendencies_nonlin()
        state_spect = self.sim.state.state_spect
        oper = self.sim.oper
        Frot_fft = tendencies_fft.get_var("rot_fft")
        rot_fft = state_spect.get_var("rot_fft")

        T_rot = (Frot_fft.conj() * rot_fft + Frot_fft * rot_fft.conj()).real / 2.0
        sum_T = oper.sum_wavenumbers(T_rot)
        self.assertAlmostEqual(sum_T, 0, places=14)
        self.sim.params.forcing.enable = True

        if mpi.nb_proc > 1:
            return

        plt.close("all")
        sim.output.spectra.plot1d()
        sim.output.spectra.plot2d()
        sim.output.spectra.load2d_mean()
        sim.output.spectra.load1d_mean()

        sim.output.spatial_means.plot()
        sim.output.spatial_means.plot_dt_energy()
        sim.output.spatial_means.plot_dt_enstrophy()
        sim.output.spatial_means.compute_time_means()
        sim.output.spatial_means.load_dataset()
        sim.output.spatial_means.time_first_saved()
        sim.output.spatial_means.time_last_saved()

        plt.close("all")

        sim.output.print_stdout.plot_energy()
        sim.output.print_stdout.plot_deltat()
        sim.output.print_stdout.plot_clock_times()

        data = sim.output.spectra_multidim.load_mean(tmin=0.2)
        spectrum = data["spectrumkykx_E"]

        assert spectrum.sum() > 1e-14
        # check NO_KY0
        assert np.allclose(spectrum[0, :].sum(), 0.0)
        # check NO_SHEAR_MODES
        assert np.allclose(spectrum[:, 0].sum(), 0.0)

        sim.output.spectra_multidim.plot()

        sim.output.spect_energy_budg.plot()
        with self.assertRaises(ValueError):
            sim.state.get_var("test")

        sim2 = fls.load_sim_for_plot(sim.output.path_run)
        sim2.output

        sim2.output.increments.load()
        sim2.output.increments.plot()
        sim2.output.increments.load_pdf_from_file()
        sim2.output.increments.plot_pdf()

        sim2.output.phys_fields.animate(
            "ux",
            dt_frame_in_sec=1e-6,
            dt_equations=0.3,
            repeat=False,
            clim=(-1, 1),
            save_file=False,
            numfig=1,
        )
        sim2.output.phys_fields.plot()
        sim2.plot_freq_diss("y")

        # `compute('q')` two times for better coverage...
        sim.state.get_var("q")
        sim.state.get_var("q")
        sim.state.get_var("div")

        path_run = sim.output.path_run

        sim3 = fls.load_state_phys_file(path_run, modif_save_params=False)
        sim3.params.time_stepping.t_end += 0.2
        sim3.time_stepping.start()

        sim3.output.phys_fields.animate(
            "ux",
            dt_frame_in_sec=1e-6,
            dt_equations=0.3,
            repeat=False,
            clim=(-1, 1),
            save_file=False,
            numfig=1,
        )
        plt.close("all")


class TestSolverNS2DInitJet(TestSimulBase):
    @classmethod
    def init_params(self):
        params = super().init_params()
        params.init_fields.type = "jet"
        params.output.HAS_TO_SAVE = False

    def test_init_jet(self):
        pass


class TestSolverNS2DInitDipole(TestSimulBase):
    @classmethod
    def init_params(self):
        params = super().init_params()
        params.init_fields.type = "dipole"
        params.output.HAS_TO_SAVE = False

    def test_init_dipole(self):
        pass


if __name__ == "__main__":
    unittest.main()



"""State class for the sw1l.modified solver
(:mod:`fluidsim.solvers.sw1l.modified.state`)
===================================================


Provides:

.. autoclass:: StateSW1LModified
   :members:
   :private-members:

"""

from fluidsim.base.setofvariables import SetOfVariables

from fluidsim.solvers.sw1l.state import StateSW1L

from fluiddyn.util import mpi


class StateSW1LModified(StateSW1L):
    """Contains the variables corresponding to the state and handles the
    access to other fields for the solver MSW1L.

    """

    @staticmethod
    def _complete_info_solver(info_solver):
        """Complete the ParamContainer info_solver.

        This is a static method!
        """
        info_solver.classes.State._set_attribs(
            {
                "keys_state_spect": ["ux_fft", "uy_fft", "eta_fft"],
                "keys_state_phys": ["ux", "uy", "eta"],
                "keys_computable": [],
                "keys_phys_needed": ["ux", "uy", "eta"],
                "keys_linear_eigenmodes": ["q_fft", "a_fft", "d_fft"],
            }
        )

    def compute(self, key, SAVE_IN_DICT=True, RAISE_ERROR=True):
        it = self.sim.time_stepping.it

        if key in self.vars_computed and it == self.it_computed[key]:
            return self.vars_computed[key]

        if key == "ux_fft":
            result = self.oper.fft2(self.state_phys.get_var("ux"))
        elif key == "uy_fft":
            result = self.oper.fft2(self.state_phys.get_var("ux"))
        elif key == "rot_fft":
            ux_fft = self.compute("ux_fft")
            uy_fft = self.compute("uy_fft")
            result = self.oper.rotfft_from_vecfft(ux_fft, uy_fft)
        elif key == "div_fft":
            ux_fft = self.compute("ux_fft")
            uy_fft = self.compute("uy_fft")
            result = self.oper.divfft_from_vecfft(ux_fft, uy_fft)
        elif key == "rot":
            rot_fft = self.compute("rot_fft")
            result = self.oper.ifft2(rot_fft)
        elif key == "div":
            div_fft = self.compute("div_fft")
            result = self.oper.ifft2(div_fft)
        elif key == "q":
            rot = self.compute("rot")
            eta = self.sim.state.state_phys.get_var("eta")
            result = rot - self.params.f * eta
        elif key == "q_fft":
            ux_fft = self.state_spect.get_var("ux_fft")
            uy_fft = self.state_spect.get_var("uy_fft")
            eta_fft = self.state_spect.get_var("eta_fft")
            rot_fft = self.oper.rotfft_from_vecfft(ux_fft, uy_fft)
            result = rot_fft - self.params.f * eta_fft
        elif key == "a_fft":
            ux_fft = self.state_spect.get_var("ux_fft")
            uy_fft = self.state_spect.get_var("uy_fft")
            eta_fft = self.state_spect.get_var("eta_fft")
            result = self.oper.afft_from_uxuyetafft(ux_fft, uy_fft, eta_fft)
        else:
            to_print = 'Do not know how to compute "' + key + '".'
            if RAISE_ERROR:
                raise ValueError(to_print)

            else:
                if mpi.rank == 0:
                    print(to_print + "\nreturn an array of zeros.")

                result = self.oper.create_arrayX(value=0.0)

        if SAVE_IN_DICT:
            self.vars_computed[key] = result
            self.it_computed[key] = it

        return result

    def init_from_uxuyfft(self, ux_fft, uy_fft):

        oper = self.oper
        ifft2 = oper.ifft2

        oper.projection_perp(ux_fft, uy_fft)
        oper.dealiasing(ux_fft, uy_fft)

        ux = ifft2(ux_fft)
        uy = ifft2(uy_fft)

        rot_fft = oper.rotfft_from_vecfft(ux_fft, uy_fft)
        rot = ifft2(rot_fft)

        eta_fft = self._etafft_no_div(ux, uy, rot)
        eta = ifft2(eta_fft)

        state_spect = self.state_spect
        state_spect.set_var("ux_fft", ux_fft)
        state_spect.set_var("uy_fft", uy_fft)
        state_spect.set_var("eta_fft", eta_fft)

        state_phys = self.state_phys
        state_phys.set_var("ux", ux)
        state_phys.set_var("uy", uy)
        state_phys.set_var("eta", eta)

    def statephys_from_statespect(self):
        """Compute the state in physical space."""
        ifft2 = self.oper.ifft2
        ux_fft = self.state_spect.get_var("ux_fft")
        uy_fft = self.state_spect.get_var("uy_fft")
        eta_fft = self.state_spect.get_var("eta_fft")
        self.state_phys.set_var("ux", ifft2(ux_fft))
        self.state_phys.set_var("uy", ifft2(uy_fft))
        self.state_phys.set_var("eta", ifft2(eta_fft))

    def return_statephys_from_statespect(self, state_spect=None):
        """Return the state in physical space."""
        ifft2 = self.oper.ifft2
        if state_spect is None:
            state_spect = self.state_spect
        ux_fft = state_spect.get_var("ux_fft")
        uy_fft = state_spect.get_var("uy_fft")
        eta_fft = state_spect.get_var("eta_fft")
        state_phys = SetOfVariables(like=self.state_phys)
        state_phys.set_var("ux", ifft2(ux_fft))
        state_phys.set_var("uy", ifft2(uy_fft))
        state_phys.set_var("eta", ifft2(eta_fft))

        return state_phys



"""Module to generate output from Basilisk (:mod:`fluidsim.base.basilisk.output`)
==================================================================================

Provides:

.. autoclass:: OutputBasilisk
   :members:
   :private-members:

"""

from ...output.base import OutputBase  # , SpecificOutput


class OutputBasilisk(OutputBase):
    @staticmethod
    def _complete_info_solver(info_solver):

        OutputBase._complete_info_solver(info_solver)

        classes = info_solver.classes.Output.classes

        base_name_mod = "fluidsim.solvers.ns2d.output"

        classes.PrintStdOut.module_name = base_name_mod + ".print_stdout"
        classes.PrintStdOut.class_name = "PrintStdOutNS2D"

        classes.PhysFields.class_name = "PhysFieldsBase2D"

    def compute_energy(self):
        """Compute the spatially averaged energy."""
        ux = self.sim.state.get_var("ux")
        uy = self.sim.state.get_var("uy")
        return 0.5 * (ux ** 2 + uy ** 2).mean()

    def compute_enstrophy(self):
        """Compute the spatially averaged enstrophy."""
        rot = self.sim.state.get_var("rot")
        return 0.5 * (rot ** 2).mean()



import unittest
from glob import glob
import os

import numpy as np

import fluiddyn as fld
import fluiddyn.util.mpi as mpi

# to get fld.show
import fluiddyn.output

from fluidsim import (
    modif_resolution_from_dir,
    load_params_simul,
    load_state_phys_file,
)

from fluidsim.util import times_start_end_from_path

from fluidsim.base.params import load_info_solver

from fluidsim.util.testing import TestSimul, classproperty, skip_if_no_fluidfft


@skip_if_no_fluidfft
class TestBaseSolverPS(TestSimul):
    @classproperty
    def Simul(cls):
        from fluidsim.base.solvers.pseudo_spect import SimulBasePseudoSpectral

        return SimulBasePseudoSpectral

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.cwd = os.getcwd()

    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()
        os.chdir(cls.cwd)

    @classmethod
    def init_params(cls):
        params = cls.params = cls.Simul.create_default_params()
        params.output.periods_plot.phys_fields = 0.2
        params.output.periods_print.print_stdout = 0.2
        params.short_name_type_run = "test_base_solver_ps"

        nh = 8
        Lh = 2 * np.pi
        params.oper.nx = nh
        params.oper.ny = nh
        params.oper.Lx = Lh
        params.oper.Ly = Lh

        params.nu_2 = 1.0

        params.time_stepping.t_end = 0.4
        params.time_stepping.type_time_scheme = "RK2"

    def test_simul(self):
        """Should be able to run a base experiment."""
        self.sim.time_stepping.start()
        load_params_simul(
            self.sim.output.path_run + "/params_simul.xml", only_mpi_rank0=False
        )

        fld.show()

        if mpi.nb_proc > 1:
            return

        modif_resolution_from_dir(
            self.sim.output.path_run, coef_modif_resol=3.0 / 2, PLOT=True
        )

        times_start_end_from_path(self.sim.output.path_run)

        path_new = os.path.join(self.sim.output.path_run, "State_phys_12x12")
        os.chdir(path_new)
        load_params_simul()
        path = glob("state_*")[0]
        load_params_simul(path)
        load_info_solver()

        sim_big = load_state_phys_file(path_new)

        for key in self.sim.state.keys_state_phys:
            var = self.sim.state.get_var(key)
            var_big = sim_big.state.get_var(key)
            assert np.mean(var ** 2) == np.mean(var_big ** 2)


if __name__ == "__main__":
    unittest.main()



"""Time stepping (:mod:`fluidsim.base.time_stepping`)
===========================================================


Provides:

.. autosummary::
   :toctree:

   base
   pseudo_spect
   finite_diff

"""



"""Preprocessing of parameters (:mod:`fluidsim.base.preprocess`)
================================================================

Provides:

.. autosummary::
   :toctree:

   base
   pseudo_spect
   .. TODO: finite_diff

"""



import numpy as np

from fluidsim.base.output import OutputBase


class Output(OutputBase):
    @staticmethod
    def _complete_info_solver(info_solver):
        """Complete the ParamContainer info_solver.

        This is a static method!
        """
        classes = info_solver.classes.Output._set_child("classes")

        base_name_mod = "fluidsim.solvers.ad1d.output"

        classes._set_child(
            "PrintStdOut",
            attribs={
                "module_name": base_name_mod + ".print_stdout",
                "class_name": "PrintStdOutAD1D",
            },
        )

        classes._set_child(
            "PhysFields",
            attribs={
                "module_name": "fluidsim.base.output.phys_fields1d",
                "class_name": "PhysFieldsBase1D",
            },
        )

    @staticmethod
    def _complete_params_with_default(params, info_solver):
        """This static method is used to complete the *params* container."""
        OutputBase._complete_params_with_default(params, info_solver)

        params.output.phys_fields.field_to_plot = "s"

    def compute_energy(self):
        return 0.5 * np.mean(self.sim.state.state_phys ** 2)



"""SW1L equations over a sphere (:mod:`fluidsim.solvers.sphere.sw1l`)
=====================================================================

.. autoclass:: InfoSolverSphereSW1L
   :members:
   :private-members:

.. autoclass:: SimulSphereSW1L
   :members:
   :private-members:

"""
from transonic import jit
from fluidsim.base.sphericalharmo.solver import (
    InfoSolverSphericalHarmo,
    SimulSphericalHarmo,
)

from fluidsim.base.setofvariables import SetOfVariables


Af = "float64[:, :]"


@jit
def compute_Frot(rot: Af, ux: Af, uy: Af, f_radial: Af):
    """Compute cross-product of absolute potential vorticity with velocity."""
    rot_abs = rot + f_radial
    F1x = rot_abs * uy
    F1y = -rot_abs * ux

    return F1x, F1y


class InfoSolverSphereSW1L(InfoSolverSphericalHarmo):
    """Contain the information on the ``sphere.sw1l`` solver."""

    def _init_root(self):
        """Init. `self` by writting the information on the solver."""

        super()._init_root()

        here = "fluidsim.solvers.sphere.sw1l"

        self.module_name = here + ".solver"
        self.class_name = "SimulSphereSW1L"
        self.short_name = "sphere.sw1l"

        self.classes.State.module_name = here + ".state"
        self.classes.State.class_name = "StateSphericalHarmoSW1L"

        self.classes.Output.module_name = here + ".output"
        self.classes.Output.class_name = "Output"


class SimulSphereSW1L(SimulSphericalHarmo):
    """Spherical-harmonics solver for shallow water equation."""

    InfoSolver = InfoSolverSphereSW1L

    @staticmethod
    def _complete_params_with_default(params):
        """Complete the `params` container (static method)."""
        SimulSphericalHarmo._complete_params_with_default(params)

        attribs = {"c2": 20}
        params._set_attribs(attribs)

    def tendencies_nonlin(self, state_spect=None, old=None):
        r"""Compute the nonlinear tendencies.

        Parameters
        ----------

        state_spect : :class:`fluidsim.base.setofvariables.SetOfVariables`
            optional

            Array containing the state, i.e. the vorticity, the divergence
            and displacement in the spectral space. If ``state_spect`` is
            provided, the variables in the physical space are computed from
            it, otherwise, they are taken from the global state of the
            simulation, ``self.state.state_phys``.

            These two possibilities are used during the time-stepping.

        old : :class:`fluidsim.base.setofvariables.SetOfVariables`
            optional

            Array containing the previous ``tendencies_sh``. This array is
            reused to save memory and improve performance.

        Returns
        -------

        tendencies_sh : :class:`fluidsim.base.setofvariables.SetOfVariables`
            An array containing the tendencies for the vorticity.

        Notes
        -----

        .. |p| mathmacro:: \partial

        The 1-layer shallow water equations are solved in the
        vector-invariant form (see section 2.2.6 Vallis 2nd edition).

        .. math::

            \p_t \hat\zeta = \hat N_\zeta - \sigma(k) \hat \zeta,
            \p_t \hat\delta = \hat N_\delta - \sigma(k) \hat \delta,
            \p_t \hat\eta = \hat N_\eta - \sigma(k) \hat \eta,

        This function computes the nonlinear term ("tendencies"). The algorithm is
        as follows,

        - Compute :math:`N_u` and :math:`N_v`, the tendencies for the velocities.
        - Take divergence and curl of the above to obtain
          :math:`N_\zeta, N_\delta`.
        - Subtract laplacian of total energy K.E. + hydrostatic pressure from
          :math:`N_\delta`.
        - Compute :math:`N_\eta = -\nabla.((\eta + 1)\mathbf{u})`

        """
        # the spherical harmonics operator
        oper = self.oper

        if state_spect is None:
            state_spect = self.state.state_spect
            state_phys = self.state.state_phys
        else:
            state_phys = self.state.return_statephys_from_statespect(state_spect)

        # get or compute rot_sh, ux and uy
        ux = state_phys.get_var("ux")
        uy = state_phys.get_var("uy")
        eta = state_phys.get_var("eta")
        rot = state_phys.get_var("rot")

        if old is None:
            tendencies_sh = SetOfVariables(like=self.state.state_spect)
        else:
            tendencies_sh = old

        Frot_sh = tendencies_sh.get_var("rot_sh")
        Fdiv_sh = tendencies_sh.get_var("div_sh")
        Feta_sh = tendencies_sh.get_var("eta_sh")
        c2 = self.params.c2

        # Absolute vorticity x Velocity
        Fux, Fuy = compute_Frot(rot, ux, uy, oper.f_radial)
        oper.divrotsh_from_vec(Fux, Fuy, Fdiv_sh, Frot_sh)

        # Subtract laplacian of K.E. + hydrostatic pressure term from
        # divergence tendency
        Fdiv_sh += oper.laplacian_sh(
            oper.sht(0.5 * (ux ** 2 + uy ** 2) + c2 * eta), negative=True
        )

        # Calculate Feta_sh = \nabla.(hu) = \nabla.((1 + \eta)u)
        oper.divrotsh_from_vec(-(eta + 1) * ux, -(eta + 1) * uy, Feta_sh)

        # oper.dealiasing(tendencies_sh)

        # def check_conservation(Fvar_sh, var_sh, var_str):
        #     import numpy as np
        #     T = np.real(Fvar_sh.conj() * var_sh + Fvar_sh * var_sh.conj()) / 2.0
        #     print(
        #         ("sum(T_{2}) = {0:9.4e} ; sum(abs(T_{2})) = {1:9.4e}").format(
        #             self.oper.sum_wavenumbers(T),
        #             self.oper.sum_wavenumbers(abs(T)),
        #             var_str
        #         ),
        #         end =" | "
        #     )
        # rot_sh = state_spect.get_var("rot_sh")
        # div_sh = state_spect.get_var("div_sh")
        # check_conservation(Frot_sh, rot_sh, "rot")
        # check_conservation(Fdiv_sh, div_sh, "div")
        # print()

        if self.params.forcing.enable:
            tendencies_sh += self.forcing.get_forcing()

        return tendencies_sh


Simul = SimulSphereSW1L


if __name__ == "__main__":

    params = Simul.create_default_params()

    params.short_name_type_run = "test"
    params.output.sub_directory = "test"

    params.nu_2 = 1e-2
    params.oper.radius = 10.0
    params.oper.omega = 0.0
    params.forcing.enable = False

    params.init_fields.type = "noise"
    params.time_stepping.deltat0 = 1e-3
    params.time_stepping.USE_CFL = False
    params.time_stepping.t_end = 10.0
    # params.time_stepping.deltat0 = 0.1

    params.output.periods_save.phys_fields = 0.25

    sim = Simul(params)
    sim.time_stepping.start()



import matplotlib.pylab as plt

from solveq2d import solveq2d


name_dir = (
    "/scratch/augier/"
    "/Results_SW1lw"
    "/Pure_standing_waves_7680x7680"
    "/SE2D_SW1lwaves_forcingw_L=50.x50._7680x7680_c=40_f=0_2013-08-06_12-26-05"
)

tmin = 194
tmax = 1000


# sim = solveq2d.load_state_phys_file(name_dir)
# sim.output.phys_fields.plot(key_field='ux')
# sim.output.phys_fields.plot(key_field='eta')


sim = solveq2d.create_sim_plot_from_dir(name_dir=name_dir)

# sim.output.print_stdout.plot()
# sim.output.spatial_means.plot()

# sim.output.spectra.plot1D(tmin=tmin, tmax=tmax, delta_t=0.,
#                           coef_compensate=2.)

# sim.output.spectra.plot2D(tmin=tmin, tmax=tmax, delta_t=0.,
#                           coef_compensate=2.)

# sim.output.spect_energy_budg.plot(tmin=tmin, tmax=tmax, delta_t=0.)

# sim.output.increments.plot(tmin=tmin, tmax=None, delta_t=0.,
#                            order=4, yscale='log')

# sim.output.prob_dens_func.plot(tmin=tmin)

sim.output.time_sigK.plot_spectra()


solveq2d.show()



import numpy as np
import matplotlib.pylab as plt

from solveq2d import solveq2d


def jump_var_delta(sim, var, deltar):
    nx = sim.param.nx
    dx = sim.param.Lx / nx
    idelta = int(round(deltar / dx))
    jump = -var
    # jump = np.zeros(var.shape)
    jump[:, 0 : nx - idelta] += var[:, idelta:nx]
    jump[:, nx - idelta : nx] += var[:, 0:idelta]
    return jump


name_dir = (
    "/scratch/augier"
    # '/home/pierre'
    "/Results_SW1lw"
    "/Pure_standing_waves_512x512/"
    # 'SE2D_SW1lwaves_forcingw_L=50.x50._512x512_c=10_f=0_2013-06-12_00-02-04'
    # 'SE2D_SW1lwaves_forcingw_L=50.x50._512x512_c=200_f=0_2013-06-12_00-39-17'
    "SE2D_SW1lwaves_forcingw_L=50.x50._512x512_c=700_f=0_2013-06-12_12-40-31"
    # '/Pure_standing_waves_1024x1024/'
    # '/Pure_standing_waves_512x512/'
    # 'SE2D_SW1lwaves_forcingw_L=50.x50._1024x1024_c=10_f=0_2013-06-12_19-34-30'
    # 'SE2D_SW1lwaves_forcingw_L=50.x50._1024x1024_c=20_f=0_2013-06-12_20-33-19'
    # 'SE2D_SW1lwaves_forcingw_L=50.x50._1024x1024_c=50_f=0_2013-06-12_21-35-00'
    # 'SE2D_SW1lwaves_forcingw_L=50.x50._1024x1024_c=100_f=0_2013-06-12_22-44-21'
    # 'SE2D_SW1lwaves_forcingw_L=50.x50._1024x1024_c=200_f=0_2013-06-13_00-33-41'
    # 'SE2D_SW1lwaves_forcingw_L=50.x50._1024x1024_c=400_f=0_2013-06-13_03-42-31'
    # 'SE2D_SW1lwaves_forcingw_L=50.x50._1024x1024_c=700_f=0_2013-06-13_09-32-03'
    # 'SE2D_SW1lwaves_forcingw_L=50.x50._1024x1024_c=1000_f=0_2013-06-13_19-31-10'
)

deltar = 1.0
divJ_limit = 1

sim = solveq2d.load_state_phys_file(t_approx=262, name_dir=name_dir)

nx = sim.param.nx
c2 = sim.param.c2
c = np.sqrt(c2)
f = sim.param.f
name_solver = sim.output.name_solver


# sim.output.phys_fields.plot(key_field='eta')
# sim.output.phys_fields.plot(key_field='div')
# sim.output.phys_fields.plot(key_field='Jx')
# sim.output.phys_fields.plot(key_field='Jy')

eta = sim.state.state_phys["eta"]
h = eta + 1

ux = sim.state.state_phys["ux"]
uy = sim.state.state_phys["uy"]

Jx = h * ux
Jy = h * uy

oper = sim.oper

Jx_fft = oper.fft2(Jx)
Jy_fft = oper.fft2(Jy)
divJ_fft = oper.divfft_from_vecfft(Jx_fft, Jy_fft)
divJ = oper.ifft2(divJ_fft)

# divJ[abs(divJ)<divJ_limit] = 0.


eta_fft = oper.fft2(eta)
px_h_fft, py_h_fft = oper.gradfft_from_fft(eta_fft)

px_h = oper.ifft2(px_h_fft)
py_h = oper.ifft2(py_h_fft)
gradh2 = px_h ** 2 + py_h ** 2

# sim.output.phys_fields.plot(field=gradh2)

gradh2_limit = 60.0 / c2

Usx = divJ * px_h / gradh2
Usy = divJ * py_h / gradh2

Usx[abs(gradh2) < gradh2_limit] = 0
Usy[abs(gradh2) < gradh2_limit] = 0

Us_over_c = np.sqrt(Usx ** 2 + Usy ** 2) / c


def limit_maxmin(var, minv, maxv):
    var[var > maxv] = maxv
    var[var < minv] = minv


maxMach = 1.5
minMach = 0.3

limit_maxmin(Us_over_c, 0.0, 2.0)

sim.output.phys_fields.plot(field=Us_over_c, key_field="Us/c", nb_contours=20)


Machx2 = (ux - Usx) ** 2 / c2
Machy2 = (uy - Usy) ** 2 / c2
Mach = np.sqrt(Machx2 + Machy2)

Mach[abs(gradh2) < gradh2_limit] = 0

limit_maxmin(Mach, minMach, maxMach)

sim.output.phys_fields.plot(
    field=Mach, key_field="local Mach number", nb_contours=20
)


x = sim.oper.x_seq
y = sim.oper.y_seq

y0 = 29.5

iy0 = np.argmin(abs(y - y0))

h0 = h[iy0]
Jx0 = Jx[iy0]
Jy0 = Jy[iy0]

# x_left_axe = 0.12
# z_bottom_axe = 0.56
# width_axe = 0.85
# height_axe = 0.37
# size_axe = [x_left_axe, z_bottom_axe,
#             width_axe, height_axe]
# fig, ax1 = sim.output.figure_axe(size_axe=size_axe)

# ax1.hold(True)
# ax1.set_xscale('linear')
# ax1.set_yscale('linear')
# ax1.set_xlabel('$x$')
# ax1.set_ylabel('$h$')
# title = ('profils, solver '+sim.output.name_solver+
# ', nh = {0:5d}'.format(nx)+
# ', c = {0:.4g}, f = {1:.4g}'.format(c, f)
# )
# ax1.set_title(title)
# ax1.hold(True)
# ax1.plot(x, h0, 'k')


# z_bottom_axe = 0.09
# size_axe[1] = z_bottom_axe
# ax2 = fig.add_axes(size_axe)
# ax2.set_xlabel('$x$')
# ax2.set_ylabel('$J_x$, $J_y$')
# ax2.plot(x, Jx0, 'r')
# ax2.plot(x, Jy0, 'b')

solveq2d.show()



"""Output (:mod:`fluidsim.solvers.models0d.predaprey.output`)
=============================================================

Provides the modules:

.. autosummary::
   :toctree:

   print_stdout

and the main output class for this solver:

.. autoclass:: Output
   :members:
   :private-members:

"""

from math import log

from fluidsim.base.output import OutputBase


class Output(OutputBase):
    """Output for prepaprey solver."""

    @staticmethod
    def _complete_info_solver(info_solver):
        """Complete the `info_solver` container (static method)."""

        OutputBase._complete_info_solver(info_solver)

        classes = info_solver.classes.Output.classes
        package = "fluidsim.solvers.models0d.predaprey.output"

        classes.PrintStdOut.module_name = package + ".print_stdout"
        classes.PrintStdOut.class_name = "PrintStdOutPredaPrey"

    @staticmethod
    def _complete_params_with_default(params, info_solver):
        """Complete the `params` container (static method)."""
        OutputBase._complete_params_with_default(params, info_solver)

        params.output.phys_fields.field_to_plot = "X"

    def compute_potential(self):
        """Compute energy(k)"""

        p = self.sim.params
        X = self.sim.state.state_phys.get_var("X")
        Y = self.sim.state.state_phys.get_var("Y")
        return p.C * log(X) - p.D * X + p.A * log(Y) - p.B * Y



"""Navier-Stokes 3D solvers (:mod:`fluidsim.solvers.ns3d`)
==========================================================

Submodules and packages
-----------------------

.. autosummary::
   :toctree:

   solver
   state
   init_fields
   forcing
   output

Derived solvers
---------------

.. autosummary::
   :toctree:

   strat
   bouss

"""



"""State for the AD1D solver (:mod:`fluidsim.solvers.ad1d.pseudo_spect.state`)
==============================================================================
"""

from fluidsim.base.state import StatePseudoSpectral

from fluiddyn.util import mpi


class StateAD1DPseudoSpectral(StatePseudoSpectral):
    """Contains the variables corresponding to the state and handles the
    access to other fields for the pseudospectral solver AD1D.

    """

    @staticmethod
    def _complete_info_solver(info_solver):
        """Complete the ParamContainer info_solver.

        This is a static method!
        """
        info_solver.classes.State._set_attribs(
            {
                "keys_state_spect": ["s_fft"],
                "keys_state_phys": ["s"],
                "keys_computable": [],
                "keys_phys_needed": ["s"],
                "keys_linear_eigenmodes": ["s"],
            }
        )

    def compute(self, key, SAVE_IN_DICT=True, RAISE_ERROR=True):
        it = self.sim.time_stepping.it
        if key in self.vars_computed and it == self.it_computed[key]:
            return self.vars_computed[key]

        if key == "dx_s":
            result = self.oper.ifft(self.compute("dx_s_fft"))
        elif key == "dx_s_fft":
            result = self.oper.pxffft_from_fft(self.state_spect.get_var("s_fft"))
        else:
            to_print = 'Do not know how to compute "' + key + '".'
            if RAISE_ERROR:
                raise ValueError(to_print)

            else:
                if mpi.rank == 0:
                    print(to_print + "\nreturn an array of zeros.")

                result = self.oper.create_arrayX(value=0.0)

        if SAVE_IN_DICT:
            self.vars_computed[key] = result
            self.it_computed[key] = it

        return result



#!/usr/bin/env python
# coding=utf8
#

from math import pi
from fluidsim.solvers.sw1l.solver import Simul
from fluidsim.util.util import mpi

params = Simul.create_default_params()

params.short_name_type_run = "no_flow"
params.init_fields.type = "constant"
params.init_fields.constant[...] = 0.0
params.output.sub_directory = "spect_energy_budg_tests_noflow"

# --------Grid parameters---------
nh = 256
Lh = 2.0 * pi
params.oper.nx = nh
params.oper.ny = nh
params.oper.Lx = Lh
params.oper.Ly = Lh
delta_x = Lh / nh

# -----Numerical method parameters----
params.f = 0.0
params.c2 = 10.0 ** 2
params.oper.coef_dealiasing = 8.0 / 9
params.time_stepping.USE_T_END = True
params.time_stepping.t_end = 5.0
params.time_stepping.it_end = 2000
params.time_stepping.USE_CFL = True
params.time_stepping.deltat0 = 1e-3

# ----- Forcing parameters -----------
params.FORCING = True
if params.FORCING:
    params.forcing.type = "waves"
    params.forcing.nkmax_forcing = 8
    params.forcing.nkmin_forcing = 5
    params.forcing.forcing_rate = 1.0
    k_max = pi / delta_x * params.oper.coef_dealiasing  # Smallest resolved scale
    k_d = k_max / 2.5
    length_scale = pi / k_d
    params.nu_8 = params.forcing.forcing_rate ** (1.0 / 3) * length_scale ** (
        22.0 / 3
    )
    # params.nu_2 = params.forcing.forcing_rate ** (1./3) * length_scale ** (4./3)

# ------Save file / StdOut parameters------------
params.output.periods_save.spatial_means = 0.1
params.output.periods_save.phys_fields = 1.0
params.output.periods_save.spectra = 0.1
params.output.periods_save.spect_energy_budg = 0.1
params.output.periods_save.increments = 0.5

params.output.periods_print.print_stdout = 0.1

# -------Plotting parameters---------------
# params.output.ONLINE_PLOT_OK = True
# params.output.spatial_means.HAS_TO_PLOT_SAVED = True
# params.output.spectra.HAS_TO_PLOT_SAVED = True

# params.output.periods_plot.phys_fields = 1.
# params.output.phys_fields.field_to_plot = 'eta'

# -------Preprocess parameters---------------
params.preprocess.enable = False
params.preprocess.init_field_scale = "unity"
params.preprocess.forcing_const = 1.0
params.preprocess.forcing_scale = "unity"
params.preprocess.viscosity_const = 2.5
params.preprocess.viscosity_scale = "forcing"
params.preprocess.viscosity_type = "hyper8"

# -------Simulation commences-------------
sim = Simul(params)
if mpi.rank == 0:
    print(
        "Froude number ~ {:3e}, Viscosity = {:3e} + {:3e}".format(
            (1.0 / params.c2), sim.params.nu_2, sim.params.nu_8
        )
    )

sim.time_stepping.start()



"""Initialization of the field (:mod:`fluidsim.solvers.ns3d.init_fields`)
=========================================================================

.. autoclass:: InitFieldsNS3D
   :members:

.. autoclass:: InitFieldsDipole
   :members:

.. autoclass:: InitFieldsNoise
   :members:

"""

import numpy as np

from fluiddyn.util import mpi

from fluidsim.base.init_fields import InitFieldsBase, SpecificInitFields


class InitFieldsDipole(SpecificInitFields):
    tag = "dipole"

    @classmethod
    def _complete_params_with_default(cls, params):
        super()._complete_params_with_default(params)

    # params.init_fields._set_child(cls.tag, attribs={'U': 1.})

    def __call__(self):
        oper = self.sim.oper
        rot2d = self.vorticity_1dipole2d()
        rot2d_fft = oper.fft2d(rot2d)

        vx2d_fft, vy2d_fft = oper.oper2d.vecfft_from_rotfft(rot2d_fft)

        vx_fft = oper.build_invariant_arrayK_from_2d_indices12X(vx2d_fft)
        vy_fft = oper.build_invariant_arrayK_from_2d_indices12X(vy2d_fft)

        fields = {"vx_fft": vx_fft, "vy_fft": vy_fft}
        self.sim.state.init_statespect_from(**fields)
        self.sim.state.statephys_from_statespect()

    def vorticity_1dipole2d(self):
        oper = self.sim.oper
        xs = oper.Lx / 2.0
        ys = oper.Ly / 2.0
        theta = np.pi / 2.3
        b = 2.5
        omega = np.zeros(oper.oper2d.shapeX_loc)

        def wz_2LO(XX, YY, b):
            return 2 * np.exp(-(XX ** 2 + (YY - (b / 2.0)) ** 2)) - 2 * np.exp(
                -(XX ** 2 + (YY + (b / 2.0)) ** 2)
            )

        XX = oper.oper2d.XX
        YY = oper.oper2d.YY

        for ip in range(-1, 2):
            for jp in range(-1, 2):
                XX_s = np.cos(theta) * (XX - xs - ip * oper.Lx) + np.sin(
                    theta
                ) * (YY - ys - jp * oper.Ly)
                YY_s = np.cos(theta) * (YY - ys - jp * oper.Ly) - np.sin(
                    theta
                ) * (XX - xs - ip * oper.Lx)
                omega += wz_2LO(XX_s, YY_s, b)
        return omega


class InitFieldsNoise(SpecificInitFields):
    """Initialize the state with noise."""

    tag = "noise"

    @classmethod
    def _complete_params_with_default(cls, params):
        """Complete the `params` container (class method)."""
        super()._complete_params_with_default(params)

        p_noise = params.init_fields._set_child(
            cls.tag, attribs={"velo_max": 1.0, "length": None}
        )
        p_noise._set_doc(
            """
velo_max: float (default 1.)

    Maximum velocity.

length: float (default 0.)

    The smallest (cutoff) scale in the noise.

"""
        )

    def __call__(self):
        vx_fft, vy_fft, vz_fft = self.compute_vv_fft()
        fields = {"vx_fft": vx_fft, "vy_fft": vy_fft, "vz_fft": vz_fft}
        params = self.sim.params
        oper = self.sim.oper

        lambda0 = params.init_fields.noise.length
        if lambda0 is None:
            lambda0 = min(oper.Lx, oper.Ly, oper.Lz) / 4.0

        def H_smooth(x, delta):
            return (1.0 + np.tanh(2 * np.pi * x / delta)) / 2.0

        k0 = 2 * np.pi / lambda0
        delta_k0 = 1.0 * k0

        K = np.sqrt(oper.K2)
        velo_max = params.init_fields.noise.velo_max

        for key in self.sim.state.keys_state_spect:
            if key not in fields:
                field = np.random.random(oper.shapeX_loc)
                field_fft = oper.fft(field)
                if mpi.rank == 0:
                    field_fft[0, 0, 0] = 0.0

                field_fft *= H_smooth(k0 - K, delta_k0)
                oper.ifft_as_arg(field_fft, field)

                value_max = np.abs(field).max()
                if mpi.nb_proc > 1:
                    value_max = oper.comm.allreduce(value_max, op=mpi.MPI.MAX)

                field_max = velo_max
                if key == "b_fft" and hasattr(params, "N"):
                    field_max *= params.N

                fields[key] = (field_max / value_max) * field_fft

        self.sim.state.init_statespect_from(**fields)
        self.sim.state.statephys_from_statespect()

    def compute_vv_fft(self):
        params_noise = self.sim.params.init_fields.noise
        return compute_solenoidal_noise_fft(
            self.sim.oper,
            params_noise.length,
            params_noise.velo_max,
        )


def compute_solenoidal_noise_fft(oper, length=None, velo_max=1, seed=42):
    """Compute a divergence-free (incompressible) random field"""

    lambda0 = length
    if lambda0 is None:
        lambda0 = oper.Lx / 4.0

    def H_smooth(x, delta):
        return (1.0 + np.tanh(2 * np.pi * x / delta)) / 2.0

    # to compute always the same field... (for 1 resolution...)
    np.random.seed(seed + mpi.rank)

    vv = [np.random.random(oper.shapeX_loc) - 0.5 for i in range(3)]

    vv_fft = []
    for ii, vi in enumerate(vv):
        vv_fft.append(oper.fft(vi))

        if mpi.rank == 0:
            vv_fft[ii][0, 0, 0] = 0.0

    oper.project_perpk3d(*vv_fft)
    oper.dealiasing(*vv_fft)

    k0 = 2 * np.pi / lambda0
    delta_k0 = 1.0 * k0

    K = np.sqrt(oper.K2)

    vv_fft = [vi_fft * H_smooth(k0 - K, delta_k0) for vi_fft in vv_fft]
    vv = [oper.ifft(ui_fft) for ui_fft in vv_fft]

    velo_max_result_random = np.sqrt(vv[0] ** 2 + vv[1] ** 2 + vv[2] ** 2).max()
    if mpi.nb_proc > 1:
        velo_max_result_random = oper.comm.allreduce(
            velo_max_result_random, op=mpi.MPI.MAX
        )

    vv = [velo_max * vi / velo_max_result_random for vi in vv]

    vv_fft = [oper.fft(vi) for vi in vv]

    return tuple(vv_fft)


class InitFieldsNS3D(InitFieldsBase):
    """Init the fields for the solver NS2D."""

    @staticmethod
    def _complete_info_solver(info_solver):
        """Complete the ParamContainer info_solver."""

        InitFieldsBase._complete_info_solver(
            info_solver, classes=[InitFieldsDipole, InitFieldsNoise]
        )



from .dimensional import Converter

from fluidsim.util.testing import TestCase


class TestDimensional(TestCase):
    def test_converter(self):

        conv = Converter(C=0.648 ** 2, h=4e-4)

        time = 10.0
        time_adim = conv.compute_time_adim(time)
        self.assertEqual(time, conv.compute_time_dim(time_adim))

        amplitude_z = 0.001
        z_adim = conv.compute_z_adim(amplitude_z)
        self.assertEqual(amplitude_z, conv.compute_z_dim(z_adim))

        amplitude_w = 0.001
        w_adim = conv.compute_w_adim(amplitude_w)
        self.assertEqual(amplitude_w, conv.compute_w_dim(w_adim))

        nu4 = 5e-7
        nu4_adim = conv.compute_nu4_adim(nu4)
        self.assertEqual(nu4, conv.compute_nu4_dim(nu4_adim))



"""Numerical simulations (:mod:`fluidsim`)
==========================================

.. _simul:

The package :mod:`fluidsim` provides an object-oriented toolkit for doing
numerical simulations of different equations (incompressible Navier-Stokes,
shallow-water, primitive equations, with and without the quasi-geostrophic limit,
adjoin equations, ...)  with different simple methods (pseudo-spectral, finite
differences) and geometries (1D, 2D and 3D periodic, 1 inhomogeneous direction,
...).

The package is organised in four sub-packages:

.. autosummary::
   :toctree:

   util
   base
   operators
   solvers
   magic

"""

from pathlib import Path
import os
import sys

if "FLUIDSIM_PATH" in os.environ:
    os.environ["TRANSONIC_DIR"] = str(
        Path(os.environ["FLUIDSIM_PATH"]) / ".transonic"
    )

_is_testing = False

if any(
    any(test_tool in arg for arg in sys.argv)
    for test_tool in ("pytest", "unittest", "fluidsim-test", "coverage")
):
    _is_testing = True
    from fluiddyn.util import mpi

    mpi.printby0(
        "Fluidsim guesses that it is tested so it "
        "loads the Agg Matplotlib backend."
    )
    import matplotlib

    matplotlib.use("Agg")
    import matplotlib.pyplot as plt

    def _show(*args, **kwargs):
        pass

    plt.show = _show

    if all(
        env_var not in os.environ
        for env_var in ("FLUID_COMPILE_CACHEDJIT", "TRANSONIC_COMPILE_JIT")
    ):
        mpi.printby0("Compilation of jit functions disabled.")
        from transonic import set_compile_jit

        set_compile_jit(False)
    elif "FLUID_COMPILE_CACHEDJIT" in os.environ:
        mpi.printby0(
            "WARNING: FLUID_COMPILE_CACHEDJIT is deprecated, use "
            "TRANSONIC_COMPILE_JIT instead."
        )

from ._version import __version__, get_local_version

from fluiddyn.io import FLUIDSIM_PATH

# has to be done before importing util
try:
    path_dir_results = Path(FLUIDSIM_PATH)
except TypeError:
    # to be able to import for transonic
    path_dir_results = None

from .util.util import (
    available_solver_keys,
    import_module_solver_from_key,
    import_simul_class_from_key,
    load_sim_for_plot,
    load_state_phys_file,
    modif_resolution_from_dir,
    modif_resolution_all_dir,
    modif_resolution_from_dir_memory_efficient,
    load_for_restart,
)

from .base.params import load_params_simul


# clean up
from . import util

try:
    from .magic import load_ipython_extension
except ImportError:
    pass

del util

# useful alias
load = load_sim_for_plot

__citation__ = r"""
@article{fluiddyn,
doi = {10.5334/jors.237},
year = {2019},
publisher = {Ubiquity Press,  Ltd.},
volume = {7},
author = {Pierre Augier and Ashwin Vishnu Mohanan and Cyrille Bonamy},
title = {{FluidDyn}: A Python Open-Source Framework for Research and Teaching in Fluid Dynamics
    by Simulations,  Experiments and Data Processing},
journal = {Journal of Open Research Software}
}

@article{fluidfft,
doi = {10.5334/jors.238},
year = {2019},
publisher = {Ubiquity Press,  Ltd.},
volume = {7},
author = {Ashwin Vishnu Mohanan and Cyrille Bonamy and Pierre Augier},
title = {{FluidFFT}: Common {API} (C$\mathplus\mathplus$ and Python)
    for Fast Fourier Transform {HPC} Libraries},
journal = {Journal of Open Research Software}
}

@article{fluidsim,
doi = {10.5334/jors.239},
year = {2019},
publisher = {Ubiquity Press,  Ltd.},
volume = {7},
author = {Mohanan, Ashwin Vishnu and Bonamy, Cyrille and Linares, Miguel Calpe and Augier, Pierre},
title = {{FluidSim}: {Modular}, {Object}-{Oriented} {Python} {Package} for
    {High}-{Performance} {CFD} {Simulations}},
journal = {Journal of Open Research Software}
}
"""

__all__ = [
    "__version__",
    "available_solver_keys",
    "get_local_version",
    "path_dir_results",
    "import_module_solver_from_key",
    "import_simul_class_from_key",
    "load_sim_for_plot",
    "load_state_phys_file",
    "modif_resolution_from_dir",
    "modif_resolution_from_dir_memory_efficient",
    "modif_resolution_all_dir",
    "load_params_simul",
    "load_for_restart",
    "__citation__",
]



"""
Plate2d InitFields (:mod:`fluidsim.solvers.plate2d.init_fields`)
======================================================================


"""

import numpy as np

from fluiddyn.util import mpi
from fluidsim.base.init_fields import InitFieldsBase, SpecificInitFields


class InitFieldsNoise(SpecificInitFields):
    tag = "noise"

    @classmethod
    def _complete_params_with_default(cls, params):
        super(cls, cls)._complete_params_with_default(params)
        params.init_fields._set_child(
            cls.tag, attribs={"velo_max": 1.0, "length": 0}
        )

    def __call__(self):
        w_fft, z_fft = self.compute_wz_fft()
        self.sim.state.init_state_from_wz_fft(w_fft, z_fft)

    def compute_wz_fft(self):
        params = self.sim.params
        oper = self.sim.oper

        lambda0 = params.init_fields.noise.length
        if lambda0 == 0:
            lambda0 = oper.Lx / 4.0

        def H_smooth(x, delta):
            return (1.0 + np.tanh(2 * np.pi * x / delta)) / 2.0

        # to compute always the same field... (for 1 resolution...)
        np.random.seed(42)  # this does not work for MPI...

        w_fft = (
            np.random.random(oper.shapeK)
            + 1j * np.random.random(oper.shapeK)
            - 0.5
            - 0.5j
        )
        z_fft = (
            np.random.random(oper.shapeK)
            + 1j * np.random.random(oper.shapeK)
            - 0.5
            - 0.5j
        )

        if mpi.rank == 0:
            w_fft[0, 0] = 0.0
            z_fft[0, 0] = 0.0

        oper.dealiasing(w_fft, z_fft)

        k0 = 2 * np.pi / lambda0
        delta_k0 = 1.0 * k0
        w_fft = w_fft * H_smooth(k0 - oper.K, delta_k0)
        z_fft = z_fft * H_smooth(k0 - oper.K, delta_k0)

        w = oper.ifft2(w_fft)
        z = oper.ifft2(z_fft)
        velo_max = np.sqrt(w ** 2 + z ** 2).max()
        if mpi.nb_proc > 1:
            velo_max = oper.comm.allreduce(velo_max, op=mpi.MPI.MAX)
        w = params.init_fields.noise.velo_max * w / velo_max
        z = params.init_fields.noise.velo_max * z / velo_max
        w_fft = oper.fft2(w)
        z_fft = oper.fft2(z)

        return w_fft, z_fft


class InitFieldsHarmonic(SpecificInitFields):
    tag = "harmonic"

    @classmethod
    def _complete_params_with_default(cls, params):
        super(cls, cls)._complete_params_with_default(params)

        params.init_fields._set_child(
            cls.tag,
            attribs={"i0": 20, "i1": 25},
        )

    def __call__(self):
        w_fft, z_fft = self.compute_wz_fft()
        self.sim.state.init_state_from_wz_fft(w_fft, z_fft)

    def compute_wz_fft(self):
        p_harmonic = self.sim.params.init_fields.harmonic
        i0 = p_harmonic.i0
        i1 = p_harmonic.i1

        oper = self.sim.oper

        w_fft = np.zeros(oper.shapeK_loc, dtype=np.complex128)
        z_fft = np.zeros(oper.shapeK_loc, dtype=np.complex128)
        w_fft[i0, i1] = 1.0
        z_fft[i0, i1] = 1.0

        w = oper.ifft2(w_fft)
        z = oper.ifft2(z_fft)

        w_fft = oper.fft2(w)
        z_fft = oper.fft2(z)

        return w_fft, z_fft


class InitFieldsPlate2D(InitFieldsBase):
    """Init the fields for the solver PLATE2D."""

    @staticmethod
    def _complete_info_solver(info_solver):
        """Complete the ParamContainer info_solver."""

        InitFieldsBase._complete_info_solver(
            info_solver, classes=[InitFieldsNoise, InitFieldsHarmonic]
        )



import unittest

from fluidsim.solvers.burgers1d.solver import Simul

from fluiddyn.util import mpi

from fluidsim.util.testing import TestSimul, skip_if_no_fluidfft


@skip_if_no_fluidfft
class TestSolverSquare1D(TestSimul):
    Simul = Simul

    @classmethod
    def init_params(cls):

        params = cls.params = cls.Simul.create_default_params()

        params.short_name_type_run = "test"
        params.output.sub_directory = "unittests"

        params.oper.nx = 40
        params.oper.Lx = 1.0

        params.time_stepping.type_time_scheme = "RK2"

        params.nu_2 = 0.01

        params.time_stepping.t_end = 0.4
        params.time_stepping.USE_CFL = False

        params.init_fields.type = "gaussian"

        params.output.periods_print.print_stdout = 0.25
        params.output.periods_save.phys_fields = 0.5
        params.output.periods_plot.phys_fields = 0.0
        params.output.phys_fields.field_to_plot = "u"

    @unittest.skipIf(
        mpi.nb_proc > 1, "MPI not implemented, for eg. sim.oper.gather_Xspace"
    )
    def test_simul(self):
        sim = self.sim
        sim.time_stepping.start()



"""Time stepping (:mod:`fluidsim.solvers.ns2d.strat.time_stepping`)
===============================================================================


Provides:

.. autoclass:: TimeSteppingPseudoSpectralStrat
   :members:
   :private-members:

"""

from math import pi
from fluiddyn.util import mpi
from fluidsim.base.time_stepping.pseudo_spect import TimeSteppingPseudoSpectral


class TimeSteppingPseudoSpectralStrat(TimeSteppingPseudoSpectral):
    """
    Class time stepping for the solver ns2d.strat.

    """

    @classmethod
    def _complete_params_with_default(cls, params):
        super()._complete_params_with_default(params)

        # Add parameter coefficient CFL GROUP VELOCITY
        params.time_stepping._set_attrib("cfl_coef_group", None)

    def _init_compute_time_step(self):
        """
        Initialization compute time step solver ns2d.strat.
        """
        super()._init_compute_time_step()

        # Coefficients dt
        self.coef_deltat_dispersion_relation = 1.0

        try:
            self.coef_group = self.params.time_stepping.cfl_coef_group
        except AttributeError:
            print("self.params.time_stepping.cfl_coef_group not in params")
            self.coef_group = 1.0

        self.coef_phase = 1.0

        has_vars = self.sim.state.has_vars
        has_ux = has_vars("ux") or has_vars("vx")
        has_uy = has_vars("uy") or has_vars("vy")
        has_b = has_vars("b")

        if has_ux and has_uy and has_b:
            self.compute_time_increment_CLF = (
                self._compute_time_increment_CFL_uxuyb
            )

        # Try to compute deltat_dispersion_relation.
        try:
            self.dispersion_relation = self.sim.compute_dispersion_relation()
        except AttributeError:
            print("compute_dispersion_relation is not implemented.")
            self.deltat_dispersion_relation = 1.0
        else:
            freq_disp_relation = self.dispersion_relation.max()

            if mpi.nb_proc > 1:
                freq_disp_relation = mpi.comm.allreduce(
                    freq_disp_relation, op=mpi.MPI.MAX
                )

            self.deltat_dispersion_relation = (
                self.coef_deltat_dispersion_relation
                * (2.0 * pi / freq_disp_relation)
            )

        # Try to compute deltat_group_vel if self.coed_group is True
        if self.coef_group:
            try:
                (
                    freq_group,
                    freq_phase,
                ) = self._compute_time_increment_group_and_phase()
            except AttributeError as e:
                print(
                    "_compute_time_increment_group_and_phase is not implemented",
                    e,
                )
                self.deltat_group_vel = 1.0
                self.deltat_phase_vel = 1.0

            else:
                if mpi.nb_proc > 1:
                    freq_group = mpi.comm.allreduce(freq_group, op=mpi.MPI.MAX)
                    freq_phase = mpi.comm.allreduce(freq_phase, op=mpi.MPI.MAX)

                self.deltat_group_vel = self.coef_group / freq_group
                self.deltat_phase_vel = self.coef_phase / freq_phase

        if self.params.forcing.enable:
            self.deltat_f = self._compute_time_increment_forcing()

    def _compute_time_increment_forcing(self):
        """
        Compute time increment of the forcing.
        """
        return 1.0 / (self.sim.params.forcing.forcing_rate ** (1.0 / 3))

    def _compute_time_increment_group_and_phase(self):
        r"""
        Compute time increment from group velocity of the internal gravity
        waves as \omega_g = max(|c_g|) \cdot max(|k|)
        """
        N = self.params.N
        oper = self.sim.oper

        KX = oper.KX
        KZ = oper.KY
        K_not0 = oper.K_not0

        # Group velocity cg
        cg_kx = (N / K_not0) * (KZ ** 2 / K_not0 ** 2)
        cg_kz = (-N / K_not0) * ((KX / K_not0) * (KZ / K_not0))
        # cg = np.sqrt(cg_kx**2 + cg_kz**2)
        max_cgx = cg_kx.max()
        max_cgz = cg_kz.max()

        freq_group = max_cgx / oper.deltax + max_cgz / oper.deltay

        # Phase velocity cp
        cp = N * (KX / K_not0 ** 2)
        max_cp = cp.max()

        freq_phase = max_cp / oper.deltax

        return freq_group, freq_phase

    def _compute_time_increment_CFL_uxuyb(self):
        """
        Compute time increment with the CFL condition solver ns2d.strat.
        """
        # Compute deltat_CFL at each time step.
        ux = self.sim.state.get_var("ux")
        uy = self.sim.state.get_var("uy")

        max_ux = abs(ux).max()
        max_uy = abs(uy).max()
        freq_CFL = max_ux / self.sim.oper.deltax + max_uy / self.sim.oper.deltay

        if mpi.nb_proc > 1:
            freq_CFL = mpi.comm.allreduce(freq_CFL, op=mpi.MPI.MAX)

        if freq_CFL > 0:
            deltat_CFL = self.CFL / freq_CFL
        else:
            deltat_CFL = self.deltat_max

        # Removed phase velocity (considered not relevant)
        if not self.coef_group:
            maybe_new_dt = min(
                deltat_CFL, self.deltat_dispersion_relation, self.deltat_max
            )
        else:
            maybe_new_dt = min(
                deltat_CFL,
                self.deltat_dispersion_relation,
                self.deltat_group_vel,
                self.deltat_max,
            )

        if self.params.forcing.enable:
            maybe_new_dt = min(maybe_new_dt, self.deltat_f)

        normalize_diff = abs(self.deltat - maybe_new_dt) / maybe_new_dt
        if normalize_diff > 0.02:
            self.deltat = maybe_new_dt



"""Base preprocess (:mod:`fluidsim.base.preprocess.base`)
=========================================================

Provides:

.. autoclass:: PreprocessBase
   :members:
   :private-members:

"""
from fluidsim_core.params import iter_complete_params


class PreprocessBase:
    _tag = "preprocess"

    @staticmethod
    def _complete_info_solver(info_solver, classes=None):
        """Complete the ParamContainer info_solver."""
        classesXML = info_solver.classes.Preprocess._set_child("classes")

        if classes is not None:
            for cls in classes:
                classesXML._set_child(
                    cls.tag,
                    attribs={
                        "module_name": cls.__module__,
                        "class_name": cls.__name__,
                    },
                )

    @staticmethod
    def _complete_params_with_default(params, info_solver):
        """This static method is used to complete the *params* container."""
        attribs = {
            "enable": False,
            "init_field_scale": "unity",
            "init_field_const": 1.0,
            "viscosity_type": "laplacian",
            "viscosity_scale": "enstrophy_forcing",
            "viscosity_const": 1.0,
            "forcing_scale": "unity",
            "forcing_const": 1.0,
        }

        params._set_child("preprocess", attribs=attribs)

        dict_classes = info_solver.classes.Preprocess.import_classes()
        iter_complete_params(params, info_solver, dict_classes.values())

    def __init__(self, sim):
        self.params = sim.params.preprocess
        self.sim = sim
        self.oper = sim.oper
        self.output = sim.output

    # dict_classes = sim.info.solver.classes.Preprocess.import_classes()

    def __call__(self):
        if self.params.enable:
            self.output.print_stdout(
                "Preprocessing initial fields, and other parameters."
            )



import unittest

import numpy as np

try:
    from fluidsht.sht2d.operators import get_simple_2d_method

    get_simple_2d_method()
    sht_avail = True
except (ImportError, NotImplementedError):
    sht_avail = False


from fluidsim.solvers.sphere.ns2d.solver import Simul
from fluidsim.util.testing import TestSimul


class TestSimulBase(TestSimul):
    Simul = Simul

    @classmethod
    def init_params(cls):

        params = cls.params = cls.Simul.create_default_params()
        params.short_name_type_run = "test"
        params.output.sub_directory = "unittests"

        params.oper.lmax = 15
        params.oper.omega = 0.0

        params.nu_8 = 2.0

        params.time_stepping.t_end = 0.5

        params.init_fields.type = "noise"

        return params


@unittest.skipUnless(sht_avail, "No SHT transform library available")
class TestSolverSphereNS2DTendency(TestSimulBase):
    @classmethod
    def init_params(self):
        params = super().init_params()

        params.output.HAS_TO_SAVE = False

    @unittest.expectedFailure
    def test_tendency(self):
        sim = self.sim
        rot_sh = sim.state.get_var("rot_sh")

        tend = sim.tendencies_nonlin(state_spect=sim.state.state_spect)
        Frot_sh = tend.get_var("rot_sh")

        T_rot = np.real(Frot_sh.conj() * rot_sh)

        ratio = sim.oper.sum_wavenumbers(T_rot) / sim.oper.sum_wavenumbers(
            abs(T_rot)
        )

        self.assertGreater(1e-15, abs(ratio))



"""Output for the ns3d solver
=============================

.. autoclass:: Output
   :members:
   :private-members:

.. autosummary::
   :toctree:

   spatial_means
   spectra
   spect_energy_budget

"""

from fluidsim.base.output import OutputBasePseudoSpectral

try:
    from fluidsim.operators.operators3d import compute_energy_from_1field
except ImportError:
    from fluidsim import _is_testing

    if not _is_testing:
        raise


class Output(OutputBasePseudoSpectral):
    @staticmethod
    def _complete_info_solver(info_solver):
        """Complete the ParamContainer info_solver."""

        OutputBasePseudoSpectral._complete_info_solver(info_solver)

        classes = info_solver.classes.Output.classes

        base_name_mod = "fluidsim.solvers.ns3d.output"

        classes.PrintStdOut.module_name = base_name_mod + ".print_stdout"
        classes.PrintStdOut.class_name = "PrintStdOutNS3D"

        classes.PhysFields.module_name = "fluidsim.base.output.phys_fields3d"
        classes.PhysFields.class_name = "PhysFieldsBase3D"

        classes._set_child(
            "Spectra",
            attribs={
                "module_name": base_name_mod + ".spectra",
                "class_name": "SpectraNS3D",
            },
        )

        classes._set_child(
            "SpatialMeans",
            attribs={
                "module_name": base_name_mod + ".spatial_means",
                "class_name": "SpatialMeansNS3D",
            },
        )

        classes._set_child(
            "SpectralEnergyBudget",
            attribs={
                "module_name": base_name_mod + ".spect_energy_budget",
                "class_name": "SpectralEnergyBudgetNS3D",
            },
        )

        classes._set_child(
            "TemporalSpectra",
            attribs={
                "module_name": "fluidsim.base.output.temporal_spectra",
                "class_name": "TemporalSpectra3D",
            },
        )

        classes._set_child(
            "SpatioTemporalSpectra",
            attribs={
                "module_name": base_name_mod + ".spatiotemporal_spectra",
                "class_name": "SpatioTemporalSpectraNS3D",
            },
        )

    @staticmethod
    def _complete_params_with_default(params, info_solver):
        """This static method is used to complete the *params* container."""
        OutputBasePseudoSpectral._complete_params_with_default(
            params, info_solver
        )
        params.output.phys_fields.field_to_plot = "rotz"

    def compute_energies_fft(self):
        vx_fft = self.sim.state.state_spect.get_var("vx_fft")
        vy_fft = self.sim.state.state_spect.get_var("vy_fft")
        vz_fft = self.sim.state.state_spect.get_var("vz_fft")
        return (
            compute_energy_from_1field(vx_fft),
            compute_energy_from_1field(vy_fft),
            compute_energy_from_1field(vz_fft),
        )

    def compute_energy_fft(self):
        nrj_x_fft, nrj_y_fft, nrj_z_fft = self.compute_energies_fft()
        return nrj_x_fft + nrj_y_fft + nrj_z_fft

    def compute_energy(self):
        energy_fft = self.compute_energy_fft()
        return self.sum_wavenumbers(energy_fft)

    def plot_summary(self, tmin=0, key_field=None):
        # pylint: disable=maybe-no-member
        self.spatial_means.plot()
        self.spectra.plot1d(tmin=tmin)
        self.spect_energy_budg.plot_fluxes(tmin=tmin)
        if key_field:
            self.phys_fields.plot(key_field)



"""Physical fields output 2d (:mod:`fluidsim.base.output.phys_fields2d`)
========================================================================

Provides:

.. autoclass:: MoviesBasePhysFields2D
   :members:
   :private-members:

.. autoclass:: PhysFieldsBase2D
   :members:
   :private-members:

"""

import os

import numpy as np

import matplotlib.pyplot as plt
from matplotlib.ticker import FormatStrFormatter

from fluiddyn.util import mpi
from .movies import MoviesBase2D
from ..params import Parameters

from .phys_fields import PhysFieldsBase


class MoviesBasePhysFields2D(MoviesBase2D):
    """Methods required to animate physical fields HDF5 files."""

    def __init__(self, output, phys_fields):
        self.phys_fields = phys_fields
        super().__init__(output)
        self._equation = None

    def init_animation(
        self, key_field, numfig, dt_equations, tmin, tmax, fig_kw, **kwargs
    ):
        """Initialize list of files and times, pcolor plot, quiver and colorbar."""
        self.phys_fields.set_of_phys_files.update_times()
        self.time_files = self.phys_fields.set_of_phys_files.times

        if dt_equations is None:
            dt_equations = np.median(np.diff(self.time_files))
            print(f"dt_equations = {dt_equations:.4f}")

        if tmax is None:
            tmax = self.time_files.max()

        super().init_animation(
            key_field, numfig, dt_equations, tmin, tmax, fig_kw, **kwargs
        )

        dt_file = (self.time_files[-1] - self.time_files[0]) / len(
            self.time_files
        )
        if dt_equations < dt_file / 4:
            raise ValueError("dt_equations < dt_file / 4")

        self._has_uxuy = self.sim.state.has_vars("ux", "uy")

        get_field_to_plot = self.phys_fields.get_field_to_plot
        field, time = get_field_to_plot(self.key_field)
        if self._has_uxuy:
            ux, time = get_field_to_plot("ux")
            uy, time = get_field_to_plot("uy")
        else:
            ux = uy = None

        INSET = True if "INSET" not in kwargs else kwargs["INSET"]
        self._init_fig(field, ux, uy, INSET, **kwargs)

        self._clim = kwargs.get("clim")
        self._set_clim()

    def _init_fig(self, field, ux=None, uy=None, INSET=True, **kwargs):
        """Initialize only the figure and related matplotlib objects. This
        method is shared by both ``animate`` and ``online_plot``
        functionalities.

        """
        self._step = step = 1 if "step" not in kwargs else kwargs["step"]
        self._QUIVER = True if "QUIVER" not in kwargs else kwargs["QUIVER"]

        x, y = self._get_axis_data(shape=field.shape)
        XX, YY = np.meshgrid(x[::step], y[::step])
        field = field[::step, ::step]

        self._im = self.ax.pcolormesh(XX, YY, field, shading="nearest")
        self._ani_cbar = self.fig.colorbar(self._im)

        self._has_uxuy = self.sim.state.has_vars("ux", "uy")

        if self._has_uxuy and self._QUIVER:
            self._ani_quiver, vmax = self.phys_fields._quiver_plot(
                self.ax, ux, uy, XX, YY
            )

        try:
            self.output.spatial_means
        except AttributeError:
            INSET = False

        if INSET and not self.sim.time_stepping.is_simul_completed():
            INSET = False

        self._ANI_INSET = INSET
        if self._ANI_INSET:
            try:
                (
                    self._ani_spatial_means_t,
                    self._ani_spatial_means_key,
                ) = self._get_spatial_means()
            except FileNotFoundError:
                print("No spatial means file => no inset plot.")
                self._ANI_INSET = False
                return

            left, bottom, width, height = [0.53, 0.67, 0.2, 0.2]
            ax2 = self.fig.add_axes([left, bottom, width, height])

            ax2.set_xlabel("t", labelpad=0.1)
            ax2.set_ylabel("E", labelpad=0.1)

            # Format of the ticks in ylabel
            ax2.yaxis.set_major_formatter(FormatStrFormatter("%.4f"))

            ax2.set_xlim(0, self._ani_spatial_means_t.max())
            # Correct visualization inset_animation 10% of the difference
            # value_max-value-min
            ax2.set_ylim(
                self._ani_spatial_means_key.min(),
                self._ani_spatial_means_key.max()
                + (
                    0.1
                    * abs(
                        self._ani_spatial_means_key.min()
                        - self._ani_spatial_means_key.max()
                    )
                ),
            )

            ax2.plot(
                self._ani_spatial_means_t,
                self._ani_spatial_means_key,
                linewidth=0.8,
                color="grey",
                alpha=0.4,
            )
            self._im_inset = ax2.plot([0], [0], color="red")

    def _get_axis_data(self, shape=None):
        """Get 1D arrays for setting the axes."""

        x, y = super()._get_axis_data()
        if shape is not None and (y.shape[0], x.shape[0]) != shape:
            path_file = os.path.join(self.output.path_run, "params_simul.xml")
            params = Parameters(path_file=path_file)
            x = np.arange(0, params.oper.Lx, params.oper.nx)
            y = np.arange(0, params.oper.Ly, params.oper.ny)

        return x, y

    def update_animation(self, frame, **fargs):
        """Loads data and updates figure."""

        print("update_animation for frame", frame, "       ", end="\r")
        time = self.ani_times[frame % len(self.ani_times)]
        step = self._step
        get_field_to_plot = self.phys_fields.get_field_to_plot

        field, time = get_field_to_plot(
            time=time,
            key=self.key_field,
            equation=self._equation,
            interpolate_time=True,
        )

        field = field[::step, ::step]

        # Update figure, quiver and colorbar
        self._im.set_array(field.flatten())

        if self._has_uxuy and self._QUIVER:
            ux, time = get_field_to_plot(
                time=time,
                key="ux",
                equation=self._equation,
                interpolate_time=True,
            )
            uy, time = get_field_to_plot(
                time=time,
                key="uy",
                equation=self._equation,
                interpolate_time=True,
            )
            vmax = np.max(np.sqrt(ux ** 2 + uy ** 2))
            skip = self.phys_fields._skip_quiver
            self._ani_quiver.set_UVC(
                ux[::skip, ::skip] / vmax, uy[::skip, ::skip] / vmax
            )
        else:
            vmax = None

        self._im.autoscale()
        self._set_clim()

        # INLET ANIMATION
        if self._ANI_INSET:
            idx_spatial = np.abs(self._ani_spatial_means_t - time).argmin()
            t = self._ani_spatial_means_t
            E = self._ani_spatial_means_key

            self._im_inset[0].set_data(t[:idx_spatial], E[:idx_spatial])

        self.phys_fields._set_title(self.ax, self.key_field, time, vmax)

    def _set_clim(self):
        """Maintains a constant colorbar throughout the animation."""

        clim = self._clim
        if clim is not None:
            self._im.set_clim(*clim)
            ticks = np.linspace(*clim, num=21, endpoint=True)
            self._ani_cbar.set_ticks(ticks)

    def _get_spatial_means(self, key_spatial="E"):
        """Get field for the inset plot."""
        # Check if key_spatial can be loaded.
        keys_spatial = ["E", "EK", "EA"]
        if key_spatial not in keys_spatial:
            raise ValueError("key_spatial not in spatial means keys.")

        # Load data for inset plot
        results = self.output.spatial_means.load()
        t = results["t"]
        E = results[key_spatial]

        return t, E


class PhysFieldsBase2D(PhysFieldsBase):
    def _init_skip_quiver(self):
        # 4% of the Lx it is a great separation between vector arrows.
        try:
            delta_quiver = 0.04 * min(self.oper.Lx, self.oper.Ly)
        except AttributeError:
            skip = 1
        else:
            skip = (
                len(self.oper.get_grid1d_seq("x")) / self.oper.Lx
            ) * delta_quiver
            skip = int(np.round(skip))
            if skip < 1:
                skip = 1
        self._skip_quiver = skip
        return skip

    def set_skip_quiver(self, skip):
        self._skip_quiver = skip

    def _init_movies(self):
        self.movies = MoviesBasePhysFields2D(self.output, self)

    def _set_title(self, ax, key, time, vmax=None):
        title = key + f", $t = {time:.3f}$"
        if vmax is not None:
            title += r", $|\vec{v}|_{max} = $" + f"{vmax:.3f}"
        ax.set_title(title + "\n" + self.output.summary_simul)

    def _init_online_plot(self):
        self.key_field = self.params.output.phys_fields.field_to_plot

        self._has_uxuy = self.sim.state.has_vars("ux", "uy")

        field, _ = self.get_field_to_plot_from_state(self.key_field)
        if self._has_uxuy:
            ux, _ = self.get_field_to_plot_from_state("ux")
            uy, _ = self.get_field_to_plot_from_state("uy")
        else:
            ux = uy = None

        if mpi.rank == 0:
            movies = self.movies
            movies.fig, movies.ax = plt.subplots()
            movies._init_fig(field, ux, uy)
            movies._im.autoscale()
            movies.fig.canvas.draw()
            plt.pause(1e-6)

    def _online_plot(self):
        """Online plot."""
        tsim = self.sim.time_stepping.t
        if tsim - self.t_last_plot >= self.period_plot:
            self.t_last_plot = tsim
            key_field = self.params.output.phys_fields.field_to_plot
            field, _ = self.get_field_to_plot_from_state(key_field)
            if self._has_uxuy:
                ux, _ = self.get_field_to_plot_from_state("ux")
                uy, _ = self.get_field_to_plot_from_state("uy")

            if mpi.rank == 0:
                # Update figure, quiver and colorbar
                self.movies._im.set_array(field.flatten())
                if self._has_uxuy:
                    vmax = np.max(np.sqrt(ux ** 2 + uy ** 2))
                    skip = self._skip_quiver
                    self.movies._ani_quiver.set_UVC(
                        ux[::skip, ::skip] / vmax, uy[::skip, ::skip] / vmax
                    )
                else:
                    vmax = None

                self._set_title(self.movies.ax, self.key_field, tsim, vmax)

                self.movies._im.autoscale()
                self.movies.fig.canvas.draw()
                plt.pause(1e-6)

    def plot(
        self,
        field=None,
        time=None,
        QUIVER=True,
        vecx="ux",
        vecy="uy",
        nb_contours=20,
        type_plot="pcolor",
        vmin=None,
        vmax=None,
        cmap=None,
        numfig=None,
        skip_quiver=None,
    ):
        """Plot a field.

        This function is surely buggy! It has to be fixed.

        Parameters
        ----------

        field : {str, np.ndarray}, optional

        time : number, optional

        QUIVER : True

        vecx : 'ux'

        vecy : 'uy'

        nb_contours : 20

        type_plot : "pcolor" or "contourf"

        vmin : None

        vmax : None

        cmap : None (usually 'viridis')

        numfig : None

        """

        is_field_ready = False

        self._has_uxuy = self.sim.state.has_vars("ux", "uy")

        key_field = None
        if field is None:
            key_field = self.get_key_field_to_plot(
                forbid_compute=time is not None
            )
        elif isinstance(field, np.ndarray):
            key_field = "given array"
            is_field_ready = True
        elif isinstance(field, str):
            key_field = field

        assert key_field is not None

        keys_state_phys = self.sim.state.keys_state_phys
        if vecx not in keys_state_phys or vecy not in keys_state_phys:
            QUIVER = False

        if (
            time is None
            and not is_field_ready
            and not self.sim.params.ONLY_COARSE_OPER
        ):
            # we have to get the field from the state
            time = self.sim.time_stepping.t
            field, _ = self.get_field_to_plot_from_state(key_field)
            if QUIVER:
                vecx, _ = self.get_field_to_plot_from_state(vecx)
                vecy, _ = self.get_field_to_plot_from_state(vecy)
        else:
            # we have to get the field from a file
            self.set_of_phys_files.update_times()
            tmax = sorted(self.set_of_phys_files.times)[-1]

            if time is None or time > tmax:
                time = sorted(self.set_of_phys_files.times)[-1]

            if key_field not in self.sim.state.keys_state_phys:
                raise ValueError("key not in state.keys_state_phys")

            field, time = self.get_field_to_plot(
                key=key_field, time=time, interpolate_time=True
            )
            if QUIVER:
                vecx, time = self.get_field_to_plot(
                    key=vecx, time=time, interpolate_time=True
                )
                vecy, time = self.get_field_to_plot(
                    key=vecy, time=time, interpolate_time=True
                )

        if mpi.rank == 0:
            if numfig is None:
                fig, ax = self.output.figure_axe()
            else:
                fig, ax = self.output.figure_axe(numfig=numfig)

            x_seq = self.oper.get_grid1d_seq("x")
            y_seq = self.oper.get_grid1d_seq("y")

            [XX_seq, YY_seq] = np.meshgrid(x_seq, y_seq)
            try:
                cmap = plt.get_cmap(cmap)
            except ValueError:
                print(
                    "Use matplotlib >= 1.5.0 for new standard colorschemes.\
                       Installed matplotlib :"
                    + plt.matplotlib.__version__
                )
                cmap = plt.get_cmap("jet")

            if type_plot == "contourf":
                contours = ax.contourf(
                    x_seq,
                    y_seq,
                    field,
                    nb_contours,
                    vmin=vmin,
                    vmax=vmax,
                    cmap=cmap,
                )
                fig.colorbar(contours)
                fig.contours = contours
            elif type_plot in ["pcolor", "pcolormesh"]:
                pc = ax.pcolormesh(
                    x_seq,
                    y_seq,
                    field,
                    shading="nearest",
                    vmin=vmin,
                    vmax=vmax,
                    cmap=cmap,
                )
                fig.colorbar(pc)
            else:
                print(f"`type_plot = {type_plot}` not implemented")
        else:
            ax = None

        if QUIVER:
            quiver, vmax = self._quiver_plot(ax, vecx, vecy, skip=skip_quiver)
        else:
            vmax = None

        if mpi.rank == 0:
            ax.set_xlabel(self.oper.axes[1])
            ax.set_ylabel(self.oper.axes[0])
            self._set_title(ax, key_field, time, vmax)

            if self.oper.Lx != self.oper.Ly:
                ax.set_aspect("equal")

            fig.tight_layout()
            fig.canvas.draw()
            plt.pause(1e-3)

    def _quiver_plot(self, ax, vecx="ux", vecy="uy", XX=None, YY=None, skip=None):
        """Superimposes a quiver plot of velocity vectors with a given axis
        object corresponding to a 2D contour plot.

        """
        if isinstance(vecx, str):
            vecx, time = self.get_field_to_plot(vecx)

        if isinstance(vecy, str):
            vecy, time = self.get_field_to_plot(vecy)

        if XX is None and YY is None:
            [XX, YY] = np.meshgrid(
                self.oper.get_grid1d_seq("x"), self.oper.get_grid1d_seq("y")
            )

        if mpi.rank == 0:
            # local variable 'normalize_diff' is assigned to but never used
            # normalize_diff = (
            #     (np.max(np.sqrt(vecx**2 + vecy**2)) -
            #      np.min(np.sqrt(vecx**2 + vecy**2))) /
            #     np.max(np.sqrt(vecx**2 + vecy**2)))
            vmax = np.max(np.sqrt(vecx ** 2 + vecy ** 2))
            # Quiver is normalized by the vmax
            # copy to avoid a bug
            if skip is None:
                skip = self._skip_quiver
            vecx_c = vecx[::skip, ::skip].copy()
            vecy_c = vecy[::skip, ::skip].copy()
            quiver = ax.quiver(
                XX[::skip, ::skip],
                YY[::skip, ::skip],
                vecx_c / vmax,
                vecy_c / vmax,
            )
        else:
            quiver = vmax = None

        return quiver, vmax
